from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Optional, TypeAlias, Union

ResourceId: TypeAlias = str


class ResourceType(Enum):
    CSV = "csv"
    JSON = "json"
    XML = "xml"
    Spreadsheet = "spreadsheet"
    NetCDF4 = "netcdf4"
    NetCDF3 = "netcdf3"
    GeoTIFF = "geotiff"
    NPDict = "np-dict"
    Shapefile = "shapefile"
    Container = "container"


@dataclass
class CSVProp:
    delimiter: str = ","


# @dataclass
# class SpreadsheetProp:
#     worksheet: Optional[str] = None


@dataclass
class Resource:
    id: ResourceId
    type: ResourceType
    prop: Optional[CSVProp] = None

    @staticmethod
    def deserialize(raw: dict):
        if raw["type"] == ResourceType.CSV.value and raw["prop"] is not None:
            prop = CSVProp(raw["prop"]["delimiter"])
        else:
            prop = None
        return Resource(raw["id"], ResourceType(raw["type"]), prop)


@dataclass
class PreprocessResourceOutput(Resource):
    original_resource_id: str = ""

    def __init__(self, resource_id: str, original_resource_id: str):
        super().__init__(id=resource_id, type=ResourceType.Container, prop=None)
        self.original_resource_id = original_resource_id

    def get_preprocessing_original_resource_id(self):
        return self.original_resource_id


class ResourceData(ABC):

    @abstractmethod
    def to_dict(self):
        pass


@dataclass
class ResourceDataFile(ResourceData):
    file: str

    def to_dict(self):
        return {"file": self.file}


@dataclass
class ResourceDataString(ResourceData):
    value: Union[str, bytes]

    def as_str(self):
        if isinstance(self.value, bytes):
            return self.value.decode()
        else:
            assert isinstance(self.value, str)
            return self.value

    def to_dict(self):
        return {
            "string": (
                self.value.decode() if isinstance(self.value, bytes) else self.value
            )
        }


@dataclass
class ResourceDataObject(ResourceData):
    value: dict | list

    def to_dict(self):
        return {"object": self.value}



from __future__ import annotations

from io import StringIO
from pathlib import Path

import serde.csv

from drepr.models.resource import ResourceDataObject, ResourceDataString


def read_source_csv(infile: Path | str | ResourceDataString | ResourceDataObject):
    if isinstance(infile, ResourceDataString):
        return serde.csv.deser(StringIO(infile.as_str()))
    elif isinstance(infile, ResourceDataObject):
        return infile.value
    return serde.csv.deser(infile)



from drepr.models.align import *
from drepr.models.attr import *
from drepr.models.drepr import *
from drepr.models.drepr_builder import *
from drepr.models.format import OutputFormat
from drepr.models.parsers.v1.resource_parser import ResourceParser
from drepr.models.path import *
from drepr.models.preprocessing import *
from drepr.models.resource import *
from drepr.models.sm import *

DEFAULT_RESOURCE_ID = ResourceParser.DEFAULT_RESOURCE_ID



from __future__ import annotations

from collections import Counter, defaultdict
from functools import cached_property
from typing import Mapping, Optional, Union

from drepr.utils.misc import get_abs_iri


class NamespaceMixin:
    prefixes: dict[str, str]

    @cached_property
    def namespace_manager(self):
        return NamespaceManager.from_prefix2ns(self.prefixes)

    @classmethod
    def is_rel_iri(cls, iri: str) -> bool:
        return iri.find("://") == -1 and iri.find(":") != -1

    def get_rel_iri(self, abs_iri: str) -> str:
        """Convert an absolute IRI to a relative IRI."""
        assert not self.is_rel_iri(abs_iri)
        prefix = self.namespace_manager.prefix_index.get(abs_iri)
        if prefix is None:
            raise ValueError(
                "Cannot create relative IRI because there is no suitable prefix"
            )
        ns = self.prefixes[prefix]
        return f"{prefix}:{abs_iri[len(ns):]}"

    def get_abs_iri(self, rel_iri: str) -> str:
        """Convert a relative IRI to an absolute IRI."""
        assert self.is_rel_iri(rel_iri), rel_iri
        return get_abs_iri(self.prefixes, rel_iri)


class NamespaceManager:
    """A helper class for converting between absolute URI and relative URI."""

    __slots__ = ("prefix2ns", "ns2prefix", "prefix_index", "prefix2len")

    def __init__(
        self,
        prefix2ns: dict[str, str],
        ns2prefix: dict[str, str],
        prefix_index: PrefixIndex,
    ):
        self.prefix2ns = prefix2ns
        self.ns2prefix = ns2prefix
        self.prefix_index = prefix_index
        self.prefix2len = {prefix: len(ns) for prefix, ns in prefix2ns.items()}

    @classmethod
    def from_prefix2ns(cls, prefix2ns: dict[str, str]):
        ns2prefix = {v: k for k, v in prefix2ns.items()}
        if len(ns2prefix) != len(prefix2ns):
            raise Exception(
                "Duplicated namespaces: %s"
                % [k for k, v in Counter(prefix2ns.values()).items() if v > 1]
            )
        prefix_index = PrefixIndex.create(ns2prefix)

        return cls(prefix2ns, ns2prefix, prefix_index)

    def normalizeUri(self, uri: str) -> str:
        prefix = self.prefix_index.get(uri)
        if prefix is None:
            return f"<{uri}>"
        return f"{prefix}:{uri[self.prefix2len[prefix]:]}"


class PrefixIndex:
    """Namespace indexing so we can quickly get prefix of a URI."""

    __slots__ = ("index", "start", "end")

    def __init__(
        self, index: dict[str, PrefixIndex | str], start: int, end: int
    ) -> None:
        self.index = index
        self.start = start
        self.end = end

    @staticmethod
    def create(ns2prefix: Mapping[str, str]):
        sorted_ns = sorted(ns2prefix.keys(), key=lambda x: len(x), reverse=True)
        if len(sorted_ns) == 0:
            raise Exception("No namespace provided")

        return PrefixIndex._create(ns2prefix, sorted_ns, 0)

    @staticmethod
    def _create(ns2prefix: Mapping[str, str], nses: list[str], start: int):
        shortest_ns = nses[-1]
        index = PrefixIndex({}, start, len(shortest_ns))

        if index.start == index.end:
            # we have an empty key, it must have more than one element because of the previous call
            index.index[""] = ns2prefix[nses[-1]]
            subindex = PrefixIndex._create(ns2prefix, nses[:-1], index.end)
            for key, node in subindex.index.items():
                index.index[key] = node
            return index

        tmp = defaultdict(list)
        for ns in nses:
            key = ns[index.start : index.end]
            tmp[key].append(ns)

        for key, lst_ns in tmp.items():
            if len(lst_ns) == 1:
                index.index[key] = ns2prefix[lst_ns[0]]
            else:
                index.index[key] = PrefixIndex._create(ns2prefix, lst_ns, index.end)
        return index

    def get(self, uri: str) -> Optional[str]:
        """Get prefix of an uri. Return None if it is not found"""
        key = uri[self.start : self.end]
        if key in self.index:
            value = self.index[key]
            if isinstance(value, PrefixIndex):
                return value.get(uri)
            return value

        if "" in self.index:
            return self.index[""]  # type: ignore

        return None

    def __str__(self):
        """Readable version of the index"""
        stack: list[tuple[int, str, Union[str, PrefixIndex]]] = list(
            reversed([(0, k, v) for k, v in self.index.items()])
        )
        out = []

        while len(stack) > 0:
            depth, key, value = stack.pop()
            indent = "    " * depth
            if isinstance(value, str):
                out.append(indent + "`" + key + "`: " + value + "\n")
            else:
                out.append(indent + "`" + key + "`:" + "\n")
                for k, v in value.index.items():
                    stack.append((depth + 1, k, v))

        return "".join(out)

    def to_dict(self):
        return {
            k: v.to_dict() if isinstance(v, PrefixIndex) else v
            for k, v in self.index.items()
        }



from __future__ import annotations

from io import StringIO
from pathlib import Path

import orjson
import serde.json

from drepr.models.resource import ResourceDataObject, ResourceDataString


def read_source_json(infile: Path | str | ResourceDataString | ResourceDataObject):
    if isinstance(infile, ResourceDataString):
        return orjson.loads(infile.as_str())
    elif isinstance(infile, ResourceDataObject):
        return infile.value
    return serde.json.deser(infile)



from collections import defaultdict
from typing import List, Optional, Union

from drepr.models.align import Alignment
from drepr.models.attr import Attr
from drepr.models.drepr import DRepr
from drepr.models.preprocessing import Preprocessing
from drepr.models.resource import Resource
from drepr.models.sm import (
    ClassNode,
    DataNode,
    DataType,
    Edge,
    LiteralNode,
    SemanticModel,
)


class DReprBuilder:
    def __init__(self):
        self.resources: List[Resource] = []
        self.preprocessing: List[Preprocessing] = []
        self.attrs: List[Attr] = []
        self.aligns: List[Alignment] = []
        self.sm: SemanticModel = SemanticModel({}, [], {})

    def add_resource(self, resource: Resource) -> "DReprBuilder":
        self.resources.append(resource)
        return self

    def add_preprocessing(self, preprocessing: Preprocessing) -> "DReprBuilder":
        self.preprocessing.append(preprocessing)
        return self

    def add_attribute(self, attr: Attr) -> "DReprBuilder":
        self.attrs.append(attr)
        return self

    def add_alignment(self, align: Alignment) -> "DReprBuilder":
        self.aligns.append(align)
        return self

    def add_sm(self) -> "SMBuilder":
        return SemanticModelBuilder(self)

    def build(self) -> DRepr:
        DRepr(self.resources, self.preprocessing, self.attrs, self.aligns, self.sm)


class SemanticModelBuilder:
    def __init__(self, repr_builder: DReprBuilder):
        self.repr_builder = repr_builder
        self.class_nodes = defaultdict(lambda: 0)
        self.sm: SemanticModel = self.repr_builder.sm

    def add_prefix(self, prefix: str, uri: str):
        self.sm.prefixes[prefix] = uri
        return self

    def add_class(self, class_name: str) -> "ClassBuilder":
        self.class_nodes[class_name] += 1
        class_id = f"{class_name}:{self.class_nodes[class_name]}"
        self.sm.nodes[class_id] = ClassNode(class_id, class_name)
        return ClassBuilder(self, class_id)

    def add_relation(self, source_class_id: str, target_class_id: str, predicate: str):
        self.sm.edges.append(Edge(source_class_id, target_class_id, predicate))
        return self

    def finish(self) -> DReprBuilder:
        return self.repr_builder


class ClassBuilder:
    def __init__(self, sm_builder: SemanticModelBuilder, class_id):
        self.sm_builder = sm_builder
        self.class_id = class_id

    def add_data_node(
        self,
        predicate: str,
        attr_id: str,
        dtype: Optional[DataType] = None,
        is_subject: bool = False,
    ):
        node_id = f"dnode:{attr_id}"
        assert node_id not in self.sm_builder.sm.nodes
        self.sm_builder.sm.nodes[node_id] = DataNode(node_id, attr_id, dtype)
        self.sm_builder.sm.edges.append(
            Edge(self.class_id, node_id, predicate, is_subject)
        )
        return self

    def add_literal_node(
        self,
        predicate: str,
        val: Union[str, int, float],
        dtype: Optional[DataType] = None,
    ):
        node_id = f"lnode:{len(self.sm_builder.sm.nodes)}"
        self.sm_builder.sm.nodes[node_id] = LiteralNode(node_id, val, dtype)
        self.sm_builder.sm.edges.append(Edge(self.class_id, node_id, predicate))
        return self

    def add_class(self, predicate: str, class_name: str):
        child_cls_builder = self.sm_builder.add_class(class_name)
        self.sm_builder.sm.edges.append(
            Edge(self.class_id, child_cls_builder.class_id, predicate)
        )
        return child_cls_builder

    def finish(self) -> SemanticModelBuilder:
        return self.sm_builder



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import List, Optional, Set, Union

from drepr.utils.misc import CacheMethod


@dataclass(frozen=True)
class Expr:
    expr: str


@dataclass
class RangeExpr:
    start: Union[int, Expr]
    end: Optional[Union[int, Expr]]  # exclusive
    step: Union[int, Expr]

    def is_select_all(self) -> bool:
        return self.start == 0 and self.end is None and self.step == 1


@dataclass
class IndexExpr:
    val: Union[str, int, Expr]
    is_optional: bool = field(
        default=False,
        metadata={"help": "Whether this index can be missing in the data source"},
    )


@dataclass
class SetIndexExpr:
    vals: Set[Union[str, int, Expr]]


class WildcardExpr(Enum):
    Values = "*"
    Names = "*~"


StepExpr = Union[RangeExpr, IndexExpr, SetIndexExpr, WildcardExpr]


@dataclass
class Path:
    steps: List[StepExpr]

    @staticmethod
    def deserialize(raw: dict) -> "Path":
        """
        Deserialize a dictionary to get back the Path object
        :param raw:
        :return:
        """
        steps = []
        for step in raw["steps"]:
            if not isinstance(step, dict):
                steps.append(WildcardExpr(step))
            elif "start" in step:
                # range index
                start = (
                    Expr(step["start"]["expr"])
                    if isinstance(step["start"], dict)
                    else step["start"]
                )
                end = (
                    Expr(step["end"]["expr"])
                    if isinstance(step["end"], dict)
                    else step["end"]
                )
                step1 = (
                    Expr(step["step"]["expr"])
                    if isinstance(step["step"], dict)
                    else step["step"]
                )
                steps.append(RangeExpr(start, end, step1))
            elif "val" in step:
                steps.append(
                    IndexExpr(
                        (
                            Expr(step["val"]["expr"])
                            if isinstance(step["val"], dict)
                            else step["val"]
                        ),
                        step["is_optional"],
                    )
                )
            elif "vals" in step:
                steps.append(
                    SetIndexExpr(
                        {
                            Expr(val["expr"]) if isinstance(val, dict) else val
                            for val in step["vals"]
                        }
                    )
                )
        return Path(steps)

    def is_step_optional(self, step_index: int):
        step = self.steps[step_index]
        return isinstance(step, IndexExpr) and step.is_optional

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def has_optional_steps(self) -> bool:
        """Check if this path has any optional steps, an optional step is a step that the key may not appear in the data source"""
        return any(isinstance(s, IndexExpr) and s.is_optional for s in self.steps)

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_optional_steps(self) -> list[int]:
        """Get the indices of optional steps in this path in sorted order"""
        return [
            i
            for i, s in enumerate(self.steps)
            if isinstance(s, IndexExpr) and s.is_optional
        ]

    def has_same_or_less_optional_steps(self, path: Path) -> bool:
        """Check if another path has the same or less optional steps as this path.

        In order for two paths to share the same optional steps, any previous steps of the optional step must be the same
        and the number of optional steps must be the same.

        In order for this path to have less optional steps than the other path, the number of optional steps must be less
        and the previous steps of the optional steps must be the same.
        """
        self_opt_steps = self.get_optional_steps()
        path_opt_steps = path.get_optional_steps()

        if self_opt_steps != path_opt_steps[: len(self_opt_steps)]:
            return False

        return (
            self.steps[: self_opt_steps[-1] + 1]
            == path.steps[: path_opt_steps[len(self_opt_steps) - 1] + 1]
        )

    def get_nary_steps(self) -> list[int]:
        """Obtain a list of indices of steps that select more than one elements"""
        unfixed_dims = []
        for d, s in enumerate(self.steps):
            if isinstance(s, (RangeExpr, SetIndexExpr, WildcardExpr)):
                unfixed_dims.append(d)

        return unfixed_dims

    def to_lang_format(self, use_json_path: bool = False) -> Union[list, str]:
        """
        Convert this Path object into the path object in the D-REPR language

        :param use_json_path: whether we should use the JSONPath or our new notation
        :return:
        """
        if use_json_path:
            jpath = ["$"]
            for step in self.steps:
                if isinstance(step, RangeExpr):
                    if any(
                        isinstance(v, Expr) for v in [step.start, step.end, step.step]
                    ):
                        raise NotImplementedError(
                            "Haven't supported JSONPath with expression yet"
                        )
                    jpath.append(f"[{step.start}:{step.end or ''}:{step.step}]")
                elif isinstance(step, IndexExpr):
                    if step.is_optional:
                        meta = "?"
                    else:
                        meta = ""
                    if isinstance(step.val, str):
                        jpath.append(f'["{step.val}"{meta}]')
                    else:
                        jpath.append(f"[{step.val}{meta}]")
                elif isinstance(step, SetIndexExpr):
                    raise NotImplementedError()
                else:
                    jpath.append(f".{step.value}")
            return "".join(jpath)

        path = []
        for step in self.steps:
            if isinstance(step, RangeExpr):
                start, end, step = [
                    (
                        ""
                        if v is None
                        else (f"${{{v.expr}}}" if isinstance(v, Expr) else v)
                    )
                    for v in [step.start, step.end, step.step]
                ]
                path.append(f"{start}..{end}:{step}")
            elif isinstance(step, IndexExpr):
                if step.is_optional:
                    path.append([step.val])
                else:
                    path.append(step.val)
            elif isinstance(step, SetIndexExpr):
                path.append(step.vals)
            elif isinstance(step, WildcardExpr):
                path.append(step.value)
            else:
                raise NotImplementedError()
        return path



import re
from abc import ABC, abstractmethod
from typing import List, Union

from drepr.models.parsers.interface import PathParser
from drepr.models.path import Expr, IndexExpr, Path, RangeExpr, WildcardExpr
from drepr.models.resource import Resource
from drepr.utils.validator import InputError


class PathParserV1(PathParser):
    """
    A path can either be a JSONPath or our list path

    1. If the path is a JSONPath, then it is a string startswith `$`. We only support the following
       type of step: range, index, list of index, and wildcard. However, wildcard is only used for selecting
       all values of an object
    2. If the path is a normal string
    """

    REG_SRANGE = re.compile(r"^(\d+)?\.\.(-?\d+)?(?::(\d+))?$")
    REG_SINDEX = re.compile(r"^(?:\$\{([^}]+)})|(\d+)|(.*)$")
    REG_SRANGE_EXPR = re.compile(
        r"^(?:(\d+)|(?:\$\{([^}]+)}))?\.\.(?:(-\d+)|(?:\$\{([^}]+)}))?(?::(\d+)|(?:\$\{([^}]+)}))?$"
    )

    REG_JPATH_BRACKET = re.compile(
        r"(?:\[(-?\d+)?\:(?:(-?\d+)(?:\:(-?\d+))?)?\])|(?:\[(-?\d+)\])|(?:\['([^']+)'\])"
    )
    REG_JPATH_DOT = re.compile(r"\.((?:(?!\.|\[).)+)")

    def parse(
        self, _resource: Resource, path: Union[str, list], parse_trace: str
    ) -> Path:
        if isinstance(path, str):
            return self.parse_jsonpath(path, parse_trace)

        if isinstance(path, list):
            return self.parse_custom_path(path, parse_trace)

        raise InputError(
            f"{parse_trace}\nERROR: the path must either be a "
            f"string (JSONPath) or a list of steps. Get {type(path)} instead"
        )

    def parse_jsonpath(self, jpath: str, parse_trace: str) -> Path:
        if not jpath.startswith("$"):
            raise InputError(
                f"{parse_trace}\nERROR: invalid json path. The path must start with `$`. "
                f"Get: {jpath}"
            )

        jpath = jpath[1:]
        steps = []
        parsing_pos = 1

        while len(jpath) > 0:
            if jpath.startswith("["):
                m = self.REG_JPATH_BRACKET.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing bracket at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                if m.group(5) is not None:
                    # match with string
                    steps.append(IndexExpr(m.group(5)))
                elif m.group(4) is not None:
                    # match with a single number
                    steps.append(IndexExpr(int(m.group(4))))
                else:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
            elif jpath.startswith(".*~"):
                # *~ select property names
                steps.append(WildcardExpr.Names)
                jpath = jpath[3:]
                parsing_pos += 3
            elif jpath.startswith(".*"):
                steps.append(WildcardExpr.Values)
                jpath = jpath[2:]
                parsing_pos += 2
            else:
                m = self.REG_JPATH_DOT.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing step at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                # after a dot, it can either be a number or a string
                if m.group(1).isdigit():
                    steps.append(IndexExpr(int(m.group(1))))
                else:
                    steps.append(IndexExpr(m.group(1)))

        return Path(steps)

    def parse_custom_path(self, path: List[str], parse_trace: str) -> Path:
        steps = []
        for i, step in enumerate(path):
            trace = f"Parsing step {i} ({step})"
            if isinstance(step, str):
                m = self.REG_SRANGE.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
                    continue

                m = self.REG_SRANGE_EXPR.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            (
                                (
                                    Expr(m.group(1)[2:-1])
                                    if m.group(1).startswith("${")
                                    else int(m.group(1))
                                )
                                if m.group(1) is not None
                                else 0
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else None
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else 1
                            ),
                        )
                    )
                    continue

                if step.startswith("${"):
                    steps.append(IndexExpr(Expr(step[2:-1])))
                else:
                    steps.append(IndexExpr(step))
            elif isinstance(step, int):
                steps.append(IndexExpr(step))
            else:
                raise InputError(
                    f"{parse_trace}\n{trace}\nERROR: step must either be string or number. Get {type(step)} instead"
                )

        return Path(steps)



from __future__ import annotations

from typing import TYPE_CHECKING

from drepr.models.parsers.v1.align_parser import AlignParser
from drepr.models.parsers.v1.attr_parser import AttrParser, ParsedAttrs
from drepr.models.parsers.v1.preprocessing_parser import PreprocessingParser
from drepr.models.parsers.v1.resource_parser import ResourceParser
from drepr.models.parsers.v2.path_parser import PathParserV2
from drepr.models.parsers.v3.sm_parser import SMParser
from drepr.models.sm import SemanticModel
from drepr.utils.validator import *

if TYPE_CHECKING:
    from drepr.models.drepr import DRepr


class ReprV3Parser:
    """
    The D-REPR language version 3 has similar to the schema of the second version.

    Difference with previous features:
    1. For spreadsheet columns, they can the letter instead of number
    2. Semantic model configuration offers more flexibility and expressiveness to create attribute/class on the fly
    """

    TOP_KEYWORDS = {
        "version",
        "resources",
        "preprocessing",
        "attributes",
        "alignments",
        "semantic_model",
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from drepr.models.drepr import DRepr

        Validator.must_be_subset(
            cls.TOP_KEYWORDS,
            raw.keys(),
            setname="Keys of D-REPR configuration",
            error_msg="Parsing D-REPR configuration",
        )

        for prop in ["version", "resources", "attributes"]:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(
            raw["version"], "3", "Parsing D-REPR configuration version"
        )
        resources = ResourceParser.parse(raw["resources"])
        attrs = ParsedAttrs()

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV2()
        preprocessing = PreprocessingParser(path_parser).parse(
            default_resource_id, resources, attrs, raw.get("preprocessing", [])
        )
        attr_parser = AttrParser(path_parser)
        attr_parser.parse(default_resource_id, resources, attrs, raw["attributes"])
        aligns = AlignParser.parse(raw.get("alignments", []))

        if "semantic_model" in raw:
            sm = SMParser(
                default_resource_id, resources, attrs, path_parser, attr_parser
            ).parse(raw["semantic_model"])
        else:
            sm = SemanticModel.get_default(attrs.attrs)

        return DRepr(resources, preprocessing, attrs.attrs, aligns, sm)

    @classmethod
    def dump(cls, drepr: "DRepr", simplify: bool = True, use_json_path: bool = False):
        raise NotImplementedError()



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from drepr.models.attr import Attr
from drepr.models.path import Path
from drepr.models.resource import Resource


@dataclass
class POutput:
    resource_id: Optional[str]
    attr: Optional[str]
    attr_path: Optional[Path]


@dataclass
class PMap:
    resource_id: str
    path: Path
    code: str
    output: Optional[POutput] = None
    change_structure: Optional[bool] = None


@dataclass
class PFilter:
    resource_id: str
    path: Path
    code: str
    output: Optional[POutput] = None


@dataclass
class PSplit:
    resource_id: str
    path: Path
    code: str
    output: Optional[POutput] = None


class RMapFunc(Enum):
    Dict2Items = "dict2items"


@dataclass
class RMap:
    resource_id: str
    path: Path
    func_id: RMapFunc
    output: Optional[POutput] = None


class PreprocessingType(Enum):
    pmap = "pmap"
    pfilter = "pfilter"
    psplit = "psplit"
    rmap = "rmap"


@dataclass
class Preprocessing:
    type: PreprocessingType
    value: Union[PMap, PFilter, PSplit, RMap]

    @staticmethod
    def deserialize(raw: dict):
        type = PreprocessingType(raw["type"])
        raw["value"]["path"] = Path.deserialize(raw["value"]["path"])
        if type == PreprocessingType.pmap:
            value = PMap(**raw["value"])
        elif type == PreprocessingType.pfilter:
            value = PFilter(**raw["value"])
        elif type == PreprocessingType.psplit:
            value = PSplit(**raw["value"])
        elif type == PreprocessingType.rmap:
            value = RMap(**raw["value"])
        else:
            raise NotImplementedError()

        return Preprocessing(type, value)

    def get_output(self) -> Optional[POutput]:
        if self.type in (
            PreprocessingType.pmap,
            PreprocessingType.pfilter,
            PreprocessingType.psplit,
            PreprocessingType.rmap,
        ):
            return self.value.output
        else:
            raise NotImplementedError(self.type)

    def is_output_new_data(self) -> bool:
        """Check if the preprocessing will generate new data. The new data is stored in a new variable"""
        if self.type == PreprocessingType.pmap:
            assert isinstance(self.value, PMap)
            return self.value.output is not None
        elif self.type == PreprocessingType.pfilter:
            assert isinstance(self.value, PFilter)
            return self.value.output is not None
        elif self.type == PreprocessingType.psplit:
            assert isinstance(self.value, PSplit)
            return self.value.output is not None
        elif self.type == PreprocessingType.rmap:
            assert isinstance(self.value, RMap)
            return self.value.output is not None
        else:
            raise NotImplementedError()

    def get_resource_id(self):
        if self.type == PreprocessingType.pmap:
            assert isinstance(self.value, PMap)
            return self.value.resource_id
        elif self.type == PreprocessingType.pfilter:
            assert isinstance(self.value, PFilter)
            return self.value.resource_id
        elif self.type == PreprocessingType.psplit:
            assert isinstance(self.value, PSplit)
            return self.value.resource_id
        elif self.type == PreprocessingType.rmap:
            assert isinstance(self.value, RMap)
            return self.value.resource_id
        else:
            raise NotImplementedError()


class Context:
    """A special instance that is accessible when user defined function is called to allow access to
    other information such as the index of the current item, or the nearby items.
    """

    def get_index(self) -> tuple:
        """Get the index of the current item"""
        raise NotImplementedError()

    def get_value(self, index: tuple):
        """Get the value of an item at a specific index"""
        raise NotImplementedError()



import copy
import re
from dataclasses import dataclass, field
from typing import Any
from uuid import uuid4

from drepr.models.attr import Attr
from drepr.models.parsers.interface import PathParser
from drepr.models.parsers.v1.attr_parser import AttrParser, ParsedAttrs
from drepr.models.resource import Resource
from drepr.models.sm import (
    DREPR_URI,
    ClassNode,
    DataNode,
    DataType,
    Edge,
    LiteralNode,
    PredefinedDataType,
    SemanticModel,
)
from drepr.utils.misc import F, get_abs_iri
from drepr.utils.namespace_mixin import NamespaceMixin
from drepr.utils.validator import InputError, Validator


@dataclass
class SMParseContext:
    nodes: dict[str, ClassNode | DataNode | LiteralNode] = field(default_factory=dict)
    edges: dict[int, Edge] = field(default_factory=dict)
    prefixes: dict[str, str] = field(default_factory=dict)


class SMParser:
    """
    This SemanticModel parser supports various ways to define a semantic model.

    ```
    semantic_model:
      <class_id>:
        [id]: <class_id>
        [uri]: <uri>
        [properties]: <properties>
        [links]: <links>
        [static_properties]: <static_properties>
        [inverse_static_properties]: <inverse_static_properties>
        [subject]: <attr>
      prefixes:
        <prefix>: <absolute uri>
    ```

    where:

    * <class_id>: (optional) is the id of the class. It can be specified via the key or as in a separate field named <id>
    * <uri> (optional): if not provided, we extract the URI from the class_id and the class id
    must have the following format: <prefix>:<name>:<alphanumeric & _> where <prefix>:<name> is
    used to create the URI.
    * <properties>: (optional)
        1. If it is a list, each item can be either:
            a. [<predicate>, <attr>]
            b. [<predicate>, <attr>, <data_type>]
            c. [<predicate>, <attr>, <is_required>]
            d. [<predicate>, <attr>, <data_type>, <is_required>]
        2. If it is a tuple, each item is (<predicate>, value) where value can be either:
            a. { attr: <attr>, [data_type: <data_type=none>], [is_required: <is_required=false>] }
            b. <attr>

        * `<predicate>` is either relative or absolute URI.
        * `<data_type>` is either relative or absolute URI specifying the data type of the property.
        * `<is_required>` is a boolean value specifying whether the property is required or not.
        * `<attr>` is the attribute of the property. It can be either a string for the attribute id or a dictionary
            to construct the attribute on fly. The dictionary will comply with the schema of the attribute.

    * <static_properties>: (optional)
        1. If it is a list, each item can be either:
            a. [<predicate>, <value>]
            b. [<predicate>, <value>, <data_type>]
        2. If it is a dictionary, each item is (<predicate>, item-value) where item-value can be either:
            a. { value: <value>, data_type: <data_type> }
            b. <value>

        * `<predicate>` is either relative or absolute URI.
        * `<data_type>` is either relative or absolute URI specifying the data type of the property.
        * `<value>` is the value of the property. It must be either a string, integer, float, or boolean.

    * <inverse_static_properties>: (optional)
        1. If it is a list, each item can be either:
            a. [<predicate>, <value>]
        2. If it is a tuple, each item is (<predicate>, item-value) where item-value must be a string, integer, float, or boolean.

    * <links>: (optional)
        1. If it is a list, each item can be either:
            a. [<predicate>, <target>]
            a. [<predicate>, <target>, <is_required>]
        2. If it is a tuple, each item is (<predicate>, item-value) where item-value can be either:
            a. { target: <target>, is_required: <is_required> }
            b. <target>
        3. If it is a dict, then it is:
            a. { prop: <predicate>, target: <target>, [is_required: <is_required=false>] }

        * `<predicate>` is either relative or absolute URI.
        * `<is_required>` is a boolean value specifying whether the link is required or not.
        * `<target>` is the target of the link. It can be either a string for the class id or a dictionary
            to construct the class on fly. The dictionary will comply with the schema of the class.

    * <subject> (optional): is the attribute that can be used as subject attribute of the class. It can be either a string
        for the attribute id or a dictionary to construct the attribute on fly. The dictionary will comply with the schema of the attribute.
    """

    CLS_KEYS = {
        "id",
        "uri",
        "properties",
        "subject",
        "links",
        "static_properties",
        "inverse_static_properties",
    }

    REG_SM_CLASS = re.compile(r"^((.+):[a-zA-Z0-9_]+)$")
    REG_SM_DNODE = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_LNODE = re.compile(
        r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$"
    )
    REG_SM_REL = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!--).)+:\d+)$")

    def __init__(
        self,
        default_resource_id: str,
        resources: list[Resource],
        attrs: ParsedAttrs,
        path_parser: PathParser,
        attr_parser: AttrParser,
    ):
        self.default_resource_id = default_resource_id
        self.resources = resources
        self.attrs = attrs
        self.path_parser = path_parser
        self.attr_parser = attr_parser

    def parse(self, sm: dict) -> SemanticModel:
        # deep copy
        sm = copy.deepcopy(sm)
        prefixes = sm.pop("prefixes", {})

        trace0 = f"Parsing `prefixes` of the semantic model"
        Validator.must_be_dict(prefixes, trace0)
        for prefix, uri in prefixes.items():
            Validator.must_be_str(uri, f"{trace0}\nParse prefix {prefix}")

        for prefix, uri in SemanticModel.get_default_prefixes().items():
            if prefix not in prefixes:
                prefixes[prefix] = uri
            elif prefixes[prefix] != uri:
                raise InputError(
                    f"{trace0}\nERROR: Prefix `{prefix}` is conflicting with predefined value `{uri}`."
                )

        context = SMParseContext(prefixes=prefixes)

        for class_id, class_conf in sm.items():
            self.parse_class(context, class_id, class_conf)

        sm = SemanticModel(context.nodes, context.edges, prefixes)
        # print(sm.nodes)
        # print(sm.edges)

        return sm

    def parse_class(
        self,
        context: SMParseContext,
        class_id: str,
        class_conf: dict,
    ):
        trace0 = f"Parsing class `{class_id}` of the semantic model"
        Validator.must_be_dict(class_conf, trace0)
        Validator.must_be_subset(
            self.CLS_KEYS, class_conf.keys(), "keys of ontology class", trace0
        )

        if "id" in class_conf:
            Validator.must_be_str(class_conf["id"], f"{trace0}\nParsing id")
            if class_id != class_conf["id"]:
                raise InputError(
                    f"{trace0}\nERROR: `id` is different from the class id. Expect to be the same"
                )

        if "uri" in class_conf:
            class_name = class_conf["uri"]
        else:
            try:
                m = self.REG_SM_CLASS.match(class_id)
                assert m is not None
                class_name = m.group(2)
            except Exception as e:
                raise InputError(
                    f"{trace0}\nERROR: invalid class_id `{class_id}`. Expect to be <string>:<alphanumeric & _>"
                )

        classnode = ClassNode(class_id, class_name)
        context.nodes[class_id] = classnode

        clsinfo_keys = [
            "properties",
            "static_properties",
            "inverse_static_properties",
            "links",
        ]
        for key in clsinfo_keys:
            values = class_conf.get(key, [])
            if isinstance(values, dict):
                values = ((k, (k, v)) for k, v in values.items())
            else:
                Validator.must_be_list(values, f"{trace0}\nParsing {key}")
                values = enumerate(values)
            for subkey, val in values:
                trace1 = f"{trace0}\nParsing {key}: {subkey}"

                if key == "properties":
                    self.parse_property(context, class_id, val, trace1)
                elif key == "static_properties":
                    self.parse_static_property(context, class_id, val, trace1)
                elif key == "inverse_static_properties":
                    self.parse_inverse_static_property(context, class_id, val, trace1)
                else:
                    assert key == "links"
                    self.parse_link(context, class_id, val, trace1)

        if "subject" in class_conf:
            trace1 = f"{trace0}\nParsing subject"
            subj_attr = self.create_attr_if_needed(class_conf["subject"], trace1)
            classnode.subject = subj_attr

    def parse_static_property(
        self,
        context: SMParseContext,
        class_id: str,
        prop: tuple | list,
        trace: str,
    ):
        """Parse a static property of a class. Can parse the following format:

        1. If prop is a list, it can be either:
            a. [<predicate>, <value>]
            b. [<predicate>, <value>, <data_type>]
        2. If prop is a tuple, it must be a tuple of (key, val), where val can be either:
            a. { value: <value>, data_type: <data_type> }
            b. <value>

        * `<predicate>` is either relative or absolute URI.
        * `<data_type>` is either relative or absolute URI specifying the data type of the property.
        * `<value>` is the value of the property. It must be either a string, integer, float, or boolean.
        """
        if isinstance(prop, list):
            if len(prop) == 2:
                predicate, value = prop
                data_type = None
            elif len(prop) == 3:
                predicate, value, data_type = prop
            else:
                raise InputError(
                    f"{trace}\nERROR: Expect each value of `static_properties` to be an array of two "
                    f"or three items (<predicate>, <value>, [data_type=none])"
                )
        else:
            if not isinstance(prop, tuple):
                raise InputError(
                    f"{trace}\nERROR: Parser Error. This is not your fault."
                )

            predicate, conf = prop
            if isinstance(conf, dict):
                Validator.must_be_subset(
                    {
                        "value",
                        "data_type",
                    },
                    conf.keys(),
                    "keys of property",
                    trace,
                )
                Validator.must_have(conf, "value", f"{trace}\n`value` is required")
                value = conf["value"]
                data_type = conf.get("data_type", None)
            else:
                value = conf
                data_type = None

        if not isinstance(value, (str, int, float, bool)):
            raise InputError(
                f"{trace}\nERROR: Value of the static property must be either a string, integer, float, or boolean"
            )
        if isinstance(value, str):
            value = str(value)
        elif isinstance(value, int):
            value = int(value)
        elif isinstance(value, float):
            value = float(value)

        if data_type is not None:
            Validator.must_be_str(data_type, f"{trace}\nParsing data type")
            try:
                data_type = DataType(data_type, context.prefixes)
            except Exception as e:
                raise InputError(f"{trace}\nERROR: {str(e)}")

        if data_type == DREPR_URI:
            assert isinstance(value, str)
            # test if the value is relative uri, if so, convert it to absolute uri
            if NamespaceMixin.is_rel_iri(value):
                value = get_abs_iri(context.prefixes, value)

        node = LiteralNode(
            node_id=f"lnode:{len(context.nodes)}", value=value, data_type=data_type
        )
        context.nodes[node.node_id] = node
        context.edges[len(context.edges)] = Edge(
            len(context.edges), class_id, node.node_id, predicate
        )

    def parse_property(
        self,
        context: SMParseContext,
        class_id: str,
        prop: tuple | list,
        trace: str,
    ):
        """Parse a property of a class. Can parse the following format:

        1. If prop is a list, it can be either:
            a. [<predicate>, <attr>]
            b. [<predicate>, <attr>, <data_type>]
            c. [<predicate>, <attr>, <is_required>]
            d. [<predicate>, <attr>, <data_type>, <is_required>]
        2. If prop is a tuple, it must be a tuple of (key, value), where value can be either:
            a. { attr: <attr>, [data_type: <data_type=none>], [is_required: <is_required=false>] }
            b. <attr>

        * `<predicate>` is either relative or absolute URI.
        * `<data_type>` is either relative or absolute URI specifying the data type of the property.
        * `<is_required>` is a boolean value specifying whether the property is required or not.
        * `<attr>` is the attribute of the property. It can be either a string for the attribute id or a dictionary
            to construct the attribute on fly. The dictionary will comply with the schema of the attribute.
        """
        if isinstance(prop, list):
            if len(prop) == 2:
                predicate, attr = prop
                data_type = None
                is_required = False
            elif len(prop) == 3:
                predicate, attr, data_type = prop
                if isinstance(data_type, bool) or data_type.lower() in {
                    "true",
                    "false",
                }:
                    is_required = (
                        data_type
                        if isinstance(data_type, bool)
                        else data_type == "true"
                    )
                    data_type = None
                else:
                    is_required = False
            elif len(prop) == 4:
                predicate, attr, data_type, is_required = prop
            else:
                raise InputError(
                    f"{trace}\nERROR: Expect each value of `properties` to be an array of two "
                    f"three or four items (<predicate>, <attribute_id>[, data_type=none][, is_required=false])"
                )
        else:
            if not isinstance(prop, tuple):
                raise InputError(
                    f"{trace}\nERROR: Parser Error. This is not your fault."
                )

            predicate, conf = prop
            if isinstance(conf, (str, list)):
                attr = conf
                data_type = None
                is_required = False
            else:
                Validator.must_be_dict(conf, trace)
                if "attr" in conf:
                    Validator.must_be_subset(
                        {
                            "attr",
                            "data_type",
                            "is_required",
                        },
                        conf.keys(),
                        "keys of property",
                        trace,
                    )
                    Validator.must_have(conf, "attr", f"{trace}\n`attr` is required")
                    attr = conf["attr"]
                    data_type = conf.get("data_type", None)
                    is_required = conf.get("is_required", False)
                else:
                    Validator.must_be_subset(
                        AttrParser.CLS_KEYS, conf.keys(), "keys of attribute", trace
                    )
                    attr = conf
                    data_type = None
                    is_required = False

        if data_type is not None:
            Validator.must_be_str(data_type, f"{trace}\nParsing data type")
            try:
                data_type = DataType(data_type, context.prefixes)
            except Exception as e:
                raise InputError(f"{trace}\nERROR: {str(e)}")

        Validator.must_be_bool(is_required, f"{trace}\nParsing is_required")

        attr_id = self.create_attr_if_needed(attr, trace)
        node = DataNode(
            node_id=f"dnode:{attr_id}", attr_id=attr_id, data_type=data_type
        )
        context.nodes[node.node_id] = node
        context.edges[len(context.edges)] = Edge(
            len(context.edges),
            class_id,
            node.node_id,
            predicate,
            is_required=is_required,
        )

    def parse_inverse_static_property(
        self,
        context: SMParseContext,
        class_id: str,
        prop: tuple | list,
        trace: str,
    ):
        """Parse an inverse static property of a class. Can parse the following format:

        1. If prop is a list, it can be either:
            a. [<predicate>, <value>]
        2. If prop is a tuple, it must be a tuple of (key, value).

        * `<predicate>` is either relative or absolute URI.
        * `<value>` is the value of the property. It must be either a string, integer, float, or boolean.
        """
        if isinstance(prop, list):
            if len(prop) == 2:
                predicate, value = prop
            else:
                raise InputError(
                    f"{trace}\nERROR: Expect each value of `inverse_static_properties` to be an array of two (<predicate>, <value>)"
                )
        else:
            if not isinstance(prop, tuple):
                raise InputError(
                    f"{trace}\nERROR: Parser Error. This is not your fault."
                )

            predicate, value = prop

        Validator.must_be_str(value, f"{trace}\nParsing value")
        if NamespaceMixin.is_rel_iri(value):
            value = get_abs_iri(context.prefixes, value)

        node = LiteralNode(
            node_id=f"lnode:{len(context.nodes)}",
            value=value,
            data_type=PredefinedDataType.drepr_uri.value,
        )
        context.nodes[node.node_id] = node
        context.edges[len(context.edges)] = Edge(
            len(context.edges), class_id, node.node_id, predicate
        )

    def parse_link(
        self,
        context: SMParseContext,
        class_id: str,
        link: tuple | list | dict,
        trace: str,
    ):
        """Parse a link of a class. Can parse the following format:

        1. If link is a list, it can be either:
            a. [<predicate>, <target>]
            a. [<predicate>, <target>, <is_required>]
        2. If link is a tuple, it must be a tuple of (key, value), where value can be either:
            a. { target: <target>, is_required: <is_required> }
            b. <target>

        * `<predicate>` is either relative or absolute URI.
        * `<is_required>` is a boolean value specifying whether the link is required or not.
        * `<target>` is the target of the link. It can be either a string for the class id or a dictionary
            to construct the class on fly. The dictionary will comply with the schema of the class.
        """
        if isinstance(link, list):
            if len(link) == 2:
                predicate, target = link
                is_required = False
            elif len(link) == 3:
                predicate, target, is_required = link
            else:
                raise InputError(
                    f"{trace}\nERROR: Expect each value of `links` to be an array of two "
                    f"or three items (<predicate>, <target>[, is_required=false])"
                )
        elif isinstance(link, dict):
            Validator.must_be_subset(
                {
                    "prop",
                    "target",
                    "is_required",
                },
                link.keys(),
                "keys of link",
                trace,
            )
            Validator.must_have(link, "prop", f"{trace}\n`prop` is required")
            Validator.must_have(link, "target", f"{trace}\n`target` is required")
            predicate = link["prop"]
            target = link["target"]
            is_required = link.get("is_required", False)
        else:
            if not isinstance(link, tuple):
                raise InputError(
                    f"{trace}\nERROR: Parser Error. This is not your fault."
                )

            predicate, conf = link
            if isinstance(conf, dict):
                if "target" in conf:
                    Validator.must_be_subset(
                        {
                            "target",
                            "is_required",
                        },
                        conf.keys(),
                        "keys of link",
                        trace,
                    )
                    Validator.must_have(
                        conf, "target", f"{trace}\n`target` is required"
                    )
                    target = conf["target"]
                    is_required = conf.get("is_required", False)
                else:
                    # construct the class on fly
                    Validator.must_be_subset(
                        self.CLS_KEYS, conf.keys(), "keys of class", trace
                    )
                    target = conf
                    is_required = False
            else:
                Validator.must_be_str(conf, f"{trace}\nParsing target")
                target = conf
                is_required = False

        Validator.must_be_bool(is_required, f"{trace}\nParsing is_required")

        if not isinstance(target, str):
            Validator.must_be_dict(target, f"{trace}\nParsing target")
            # construct the class on fly
            if "id" in target:
                target_class_id = target["id"]
            else:
                target_class_id = f'{target["uri"]}:{len(context.nodes)}'

            self.parse_class(context, target_class_id, target)
        else:
            target_class_id = target

        context.edges[len(context.edges)] = Edge(
            len(context.edges),
            class_id,
            target_class_id,
            predicate,
            is_required=is_required,
        )

    def create_attr_if_needed(self, attr: Any, trace: str):
        if isinstance(attr, str) and attr[0] != "$":
            if not self.attrs.has_been_reference_before(attr):
                raise InputError(
                    f"{trace}\nAttribute {attr} is not defined yet. If you want to create it on fly with only `path`, please use list format for the attribute path"
                )
            return attr

        # construct the attribute on fly
        if isinstance(attr, str):
            # attribute path -- only work if only a single resource
            if len(self.resources) > 1:
                raise InputError(
                    f"{trace}\nERROR: Cannot create an attribute only from path in a multi-resource environment"
                )

            path = self.path_parser.parse(self.resources[0], attr, trace)
            newattr = Attr(
                str(uuid4()).replace("-", "_"),
                self.default_resource_id,
                path,
                missing_values=[],
            )
        else:
            trace1 = f"{trace}\nParsing attribute"
            Validator.must_be_dict(attr, trace1)
            Validator.must_have(attr, "id", trace1)
            Validator.must_be_str(attr["id"], f"{trace1}\nParsing attribute id")

            attr_id = attr["id"]
            if self.attrs.has_been_reference_before(attr_id):
                raise InputError(f"{trace1}\nERROR: Duplicated attribute id: {attr_id}")
            newattr = self.attr_parser.parse_expanded_def(
                self.default_resource_id, self.resources, attr_id, attr, trace1
            )

        self.attrs.add(newattr)
        return newattr.id



from __future__ import annotations

from typing import Any, List, Type, Union

from drepr.models.parsers.v1.attr_parser import ParsedAttrs
from drepr.models.parsers.v1.path_parser import PathParser
from drepr.models.path import Path
from drepr.models.preprocessing import (
    PFilter,
    PMap,
    POutput,
    Preprocessing,
    PreprocessingType,
    PSplit,
    RMap,
)
from drepr.models.resource import Resource
from drepr.utils.validator import Validator


class PreprocessingParser:
    """
    Preprocessing are defined as a list

    ```
    - type: <preprocessing_type>
      # other properties
    ```

    1. If <preprocessing_type> is `pmap`, its other properties are:

        ```
        - type: pmap
          [resource_id]: <resource_id>
          path: <path>
          [output]: <output> (default is None)
          [change_structure]: null|true|false (default is null)
          code: str
        ```

    2. If <preprocessing_type> is `pfilter`, its properties are:

        ```
        - type: pfilter
          [resource_id]: <resource_id>
          path: <path>
          [output]: <output> (default is None)
          code: str
        ```

    3. If <preprocessing_type> is `rmap`, its properties are:

        ```
        - type: rmap
          resource_id: <resource_id>
          path: <path>
          func_id: <func_id>
          [output]: <output> (default is None)
        ```

    where `output` is either <resource_id> or an object of
        resource_id: <resource_id>
        attr: <attr_id>
        [attr_path]: <path>

    """

    PREPRO_TYPES = {x.value for x in PreprocessingType}

    def __init__(self, path_parser: PathParser):
        self.path_parser = path_parser

    def parse(
        self,
        default_resource_id: str,
        resources: List[Resource],
        attrs: ParsedAttrs,
        conf: list,
    ) -> List[Preprocessing]:
        Validator.must_be_list(conf, "Parsing preprocessing")
        result: list[Preprocessing] = []

        for i, prepro in enumerate(conf):
            trace0 = f"Parsing preprocessing at position {i}"
            Validator.must_be_dict(prepro, trace0)
            Validator.must_have(prepro, "type", trace0)
            Validator.must_in(
                prepro["type"], self.PREPRO_TYPES, f"{trace0}\nParsing property `type`"
            )
            prepro_type = PreprocessingType(prepro["type"])

            if "resource_id" in prepro:
                Validator.must_be_str(
                    prepro["resource_id"], f"{trace0}\nParsing property `resource_id`"
                )
                resource_id = prepro["resource_id"]
            else:
                resource_id = default_resource_id

            trace1 = f"{trace0}\nParsing property `path`"
            Validator.must_have(prepro, "path", trace1)
            path = self.path_parser.parse(
                self.path_parser.get_resource(resources, resource_id, trace0),
                prepro["path"],
                trace1,
            )

            if prepro_type == PreprocessingType.pmap:
                value = self.parse_pmap(resource_id, path, prepro, trace0)
            elif prepro_type == PreprocessingType.pfilter:
                value = self.parse_pfilter_psplit(
                    resource_id, path, prepro, trace0, PFilter
                )
            elif prepro_type == PreprocessingType.psplit:
                value = self.parse_pfilter_psplit(
                    resource_id, path, prepro, trace0, PSplit
                )
            elif prepro_type == PreprocessingType.rmap:
                value = self.parse_rmap(resource_id, path, prepro, trace0)
            else:
                raise NotImplemented(
                    f"Not implement the parser for preprocessing function with type {prepro_type}"
                )

            if value.output is not None and value.output.attr is not None:
                attrs.add_preprocessing_attr(value.output.attr)
            result.append(Preprocessing(prepro_type, value))

        return result

    def parse_pmap(
        self, resource_id: str, path: Path, prepro: dict, trace0: str
    ) -> PMap:
        trace1 = f"{trace0}\nParsing property `code`"
        Validator.must_have(prepro, "code", trace1)
        Validator.must_be_str(prepro["code"], trace1)
        code = prepro["code"]

        if "output" in prepro and prepro["output"] is not None:
            trace1 = f"{trace0}\nParsing property `output`"
            output = self.parse_output(prepro["output"], trace1)
        else:
            output = None

        if "change_structure" in prepro and prepro["change_structure"] is not None:
            trace1 = f"{trace0}\nParsing property `change_structure`"
            Validator.must_be_bool(prepro["change_structure"], trace1)
            change_structure = prepro["change_structure"]
        else:
            change_structure = None

        return PMap(resource_id, path, code, output, change_structure)

    def parse_pfilter_psplit(
        self,
        resource_id: str,
        path: Path,
        prepro: dict,
        trace0: str,
        cls: Union[Type[PFilter], Type[PSplit]],
    ) -> PFilter | PSplit:
        trace1 = f"{trace0}\nParsing property `code`"
        Validator.must_have(prepro, "code", trace1)
        Validator.must_be_str(prepro["code"], trace1)
        code = prepro["code"]

        if "output" in prepro:
            trace1 = f"{trace0}\nParsing property `output`"
            output = self.parse_output(prepro["output"], trace1)
        else:
            output = None

        return cls(resource_id, path, code, output)

    def parse_rmap(
        self, resource_id: str, path: Path, prepro: dict, trace0: str
    ) -> RMap:
        trace1 = f"{trace0}\nParsing property `func_id`"
        Validator.must_have(prepro, "func_id", trace1)
        Validator.must_be_str(prepro["func_id"], trace1)
        func_id = prepro["func_id"]

        if "output" in prepro:
            trace1 = f"{trace0}\nParsing property `output`"
            output = self.parse_output(prepro["output"], trace1)
        else:
            output = None

        return RMap(resource_id, path, func_id, output)

    def parse_output(self, output: Any, trace: str) -> POutput:
        if isinstance(output, str):
            return POutput(resource_id=output, attr=None, attr_path=None)
        else:
            Validator.must_be_dict(
                output, f"{trace}\nParsing output. Must be either string or dictionary"
            )
            Validator.must_be_subset(
                {"resource_id", "attr", "attr_path"},
                output.keys(),
                "keys of preprocessing output",
                trace,
            )

            if "resource_id" in output:
                Validator.must_be_str(
                    output["resource_id"], f"{trace}\nParsing output's resource id"
                )

            attr_path = None
            if "attr_path" in output:
                attr_path = self.path_parser.parse(
                    resource=None,
                    path=output["attr_path"],
                    parse_trace=f"{trace}\nParsing output's attr_path",
                )

            if "attr" in output:
                Validator.must_be_str(
                    output["attr"], f"{trace}\nParsing output's attribute"
                )

            return POutput(
                resource_id=output.get("resource_id"),
                attr=output.get("attr"),
                attr_path=attr_path,
            )



from __future__ import annotations

from collections import namedtuple

ResourceKey = namedtuple("ResourceKey", ["resource"])


class VarSpace:
    output_file = lambda: ("output_file",)
    resource = lambda resource: ResourceKey(
        f"resource={resource}",
    )
    resource_data = lambda resource: (f"resource-data={resource}",)
    preprocessing_udf = lambda func_name: (func_name,)
    preprocessing_udf_value = lambda preprocessing_id: (
        f"preprocessing-id={preprocessing_id}",
        "udf-value",
    )
    preprocessing_udf_context = lambda preprocessing_id: (
        f"preprocessing-id={preprocessing_id}",
        "udf-context",
    )
    preprocessing_output = lambda preprocessing_id: (
        f"preprocessing-id={preprocessing_id}",
        "output",
    )
    writer = lambda: ("writer",)
    attr_index_dim = lambda resource, attr, di: (
        f"resource={resource}",
        f"attr={attr}",
        f"index-dim={di}",
    )
    attr_value_dim = lambda resource, attr, di: (
        f"resource={resource}",
        f"attr={attr}",
        f"value-dim={di}",
    )
    attr_values_dim = lambda resource, attr, di: (
        f"resource={resource}",
        f"attr={attr}",
        f"values-dim={di}",
    )
    attr_missing_values = lambda attr: (f"attr-missing-values={attr}",)
    has_attr_value_dim = lambda resource, attr, di: (
        f"resource={resource}",
        f"attr={attr}",
        f"has-value-dim={di}",
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Union

from drepr.models.path import Path
from drepr.models.resource import Resource
from drepr.utils.validator import InputError


class PathParser(ABC):
    @abstractmethod
    def parse(
        self, resource: Optional[Resource], path: Union[str, list], parse_trace: str
    ) -> Path:
        pass

    # noinspection PyMethodMayBeStatic
    def get_resource(
        self, resources: List[Resource], resource_id: str, trace: str
    ) -> Resource:
        for res in resources:
            if res.id == resource_id:
                return res
        raise InputError(
            f"{trace}\nERROR: Refer to path of an nonexistent resource: {resource_id}"
        )



from collections import defaultdict
from dataclasses import asdict
from typing import TYPE_CHECKING, List

from drepr.models.align import AlignmentType, RangeAlignment
from drepr.models.parsers.v1.align_parser import AlignParser
from drepr.models.parsers.v1.attr_parser import AttrParser, ParsedAttrs
from drepr.models.parsers.v1.path_parser import PathParserV1
from drepr.models.parsers.v1.preprocessing_parser import PreprocessingParser
from drepr.models.parsers.v1.resource_parser import ResourceParser
from drepr.models.parsers.v1.sm_parser import SMParser
from drepr.models.resource import Resource
from drepr.models.sm import ClassNode, DataNode, LiteralNode, SemanticModel
from drepr.utils.validator import *

if TYPE_CHECKING:
    from drepr.models.drepr import DRepr


class ReprV1Parser:
    """
    The DREPR language version 1 has the following schema:

    ```
    version: '1'
    resources: <resources>
    [preprocessing]: <preprocessing> (default is empty list)
    attributes: <attributes>
    [alignments]: <alignments> (default is empty list)
    semantic_model: <semantic_model>
    ```
    """

    TOP_KEYWORDS = {
        "version",
        "resources",
        "preprocessing",
        "attributes",
        "alignments",
        "semantic_model",
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from drepr.models.drepr import DRepr

        Validator.must_be_subset(
            cls.TOP_KEYWORDS,
            raw.keys(),
            setname="Keys of D-REPR configuration",
            error_msg="Parsing D-REPR configuration",
        )

        for prop in ["version", "resources", "attributes"]:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(
            raw["version"], "1", "Parsing D-REPR configuration version"
        )
        resources = ResourceParser.parse(raw["resources"])
        attrs = ParsedAttrs()

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV1()

        preprocessing = PreprocessingParser(path_parser).parse(
            default_resource_id, resources, attrs, raw.get("preprocessing", [])
        )
        AttrParser(path_parser).parse(
            default_resource_id, resources, attrs, raw["attributes"]
        )
        aligns = AlignParser.parse(raw.get("alignments", []))

        if "semantic_model" in raw:
            sm = SMParser.parse(raw["semantic_model"])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = SemanticModel.get_default(attrs.attrs)

        return DRepr(resources, preprocessing, attrs.attrs, aligns, sm)

    @classmethod
    def dump(cls, drepr: "DRepr", simplify: bool = True, use_json_path: bool = False):
        version = "1"
        sm = OrderedDict(
            [
                ("data_nodes", OrderedDict()),
                ("relations", []),
                ("literal_nodes", []),
                ("subjects", OrderedDict([])),
                ("prefixes", drepr.sm.prefixes),
            ]
        )

        class_ids: Dict[str, Dict[str, str]] = defaultdict(lambda: {})
        for node in drepr.sm.nodes.values():
            if isinstance(node, ClassNode):
                class_ids[node.label][
                    node.node_id
                ] = f"{node.label}:{len(class_ids[node.label]) + 1}"

        for node in drepr.sm.nodes.values():
            if isinstance(node, DataNode):
                edge = [
                    e for e in drepr.sm.edges.values() if e.target_id == node.node_id
                ][0]
                sm["data_nodes"][
                    node.attr_id
                ] = f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}"
                if node.data_type is not None:
                    sm["data_nodes"][node.attr_id] += f"^^{node.data_type.value}"

            if isinstance(node, LiteralNode):
                edge = [e for e in drepr.sm.edges if e.target_id == node.node_id][0]
                sm["literal_nodes"].append(
                    f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}--{node.value}"
                )
                if node.data_type is not None:
                    sm["literal_nodes"][-1] += f"^^{node.data_type.value}"

        for edge in drepr.sm.edges.values():
            if isinstance(drepr.sm.nodes[edge.source_id], ClassNode) and isinstance(
                drepr.sm.nodes[edge.target_id], ClassNode
            ):
                sm["relations"].append(
                    f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}--{class_ids[drepr.sm.nodes[edge.target_id].label][edge.target_id]}"
                )

            if edge.is_subject:
                sm["subjects"][
                    class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]
                ] = drepr.sm.nodes[edge.target_id].attr_id

        preprocessing: List[dict] = []
        for prepro in drepr.preprocessing:
            preprocessing.append(OrderedDict([("type", prepro.type.value)]))
            for k, v in asdict(prepro.value).items():
                preprocessing[-1][k] = v
            preprocessing[-1]["path"] = prepro.value.path.to_lang_format(use_json_path)

        return OrderedDict(
            [
                ("version", version),
                (
                    "resources",
                    OrderedDict(
                        [
                            (
                                res.id,
                                OrderedDict(
                                    [("type", res.type.value)]
                                    + (
                                        [(k, v) for k, v in asdict(res.prop).items()]
                                        if res.prop is not None
                                        else []
                                    )
                                ),
                            )
                            for res in drepr.resources
                        ]
                    ),
                ),
                ("preprocessing", preprocessing),
                (
                    "attributes",
                    OrderedDict(
                        [
                            (
                                attr.id,
                                OrderedDict(
                                    [
                                        ("resource_id", attr.resource_id),
                                        (
                                            "path",
                                            attr.path.to_lang_format(use_json_path),
                                        ),
                                        ("unique", attr.unique),
                                        ("sorted", attr.sorted.value),
                                        ("value_type", attr.value_type.value),
                                        ("missing_values", attr.missing_values),
                                    ]
                                ),
                            )
                            for attr in drepr.attrs
                        ]
                    ),
                ),
                (
                    "alignments",
                    [
                        (
                            OrderedDict(
                                [
                                    ("type", AlignmentType.Range.value),
                                    ("source", align.source),
                                    ("target", align.target),
                                    (
                                        "aligned_dims",
                                        [
                                            OrderedDict(
                                                [
                                                    ("source", step.source_idx),
                                                    ("target", step.target_idx),
                                                ]
                                            )
                                            for step in align.aligned_steps
                                        ],
                                    ),
                                ]
                            )
                            if isinstance(align, RangeAlignment)
                            else OrderedDict(
                                [
                                    ("type", AlignmentType.Value.value),
                                    ("source", align.source),
                                    ("target", align.target),
                                ]
                            )
                        )
                        for align in drepr.aligns
                    ],
                ),
                ("semantic_model", sm),
            ]
        )



from __future__ import annotations

from collections import OrderedDict
from dataclasses import asdict, dataclass
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Any, Optional, Union

import orjson
from ruamel.yaml import YAML

from drepr.models.align import (
    AlignedStep,
    Alignment,
    AlignmentType,
    AutoAlignment,
    IdenticalAlign,
    RangeAlignment,
    ValueAlignment,
)
from drepr.models.attr import Attr
from drepr.models.parsers.v1 import ReprV1Parser
from drepr.models.parsers.v2 import ReprV2Parser
from drepr.models.parsers.v3 import ReprV3Parser
from drepr.models.preprocessing import PFilter, PMap, Preprocessing, PSplit, RMap
from drepr.models.resource import CSVProp, Resource, ResourceType
from drepr.models.sm import ClassNode, DataNode, LiteralNode, SemanticModel
from drepr.utils.validator import InputError, Validator

yaml = YAML()
yaml.Representer.add_representer(OrderedDict, yaml.Representer.represent_dict)


@dataclass
class EngineFormat:
    model: dict[str, Any]
    edges_optional: list[bool]
    resource_idmap: dict[str, int]
    attribute_idmap: dict[str, int]
    sm_node_idmap: dict[str, int]


@dataclass
class DRepr:
    resources: list[Resource]
    preprocessing: list[Preprocessing]
    attrs: list[Attr]
    aligns: list[Alignment]
    sm: SemanticModel

    @staticmethod
    def parse(raw: dict) -> "DRepr":
        Validator.must_have(raw, "version", "Parsing D-REPR configuration")
        raw["version"] = str(raw["version"])
        if raw["version"] == "1":
            model = ReprV1Parser.parse(raw)
            model.assert_valid()
            return model
        elif raw["version"] == "2":
            model = ReprV2Parser.parse(raw)
            model.assert_valid()
            return model
        elif raw["version"] == "3":
            model = ReprV3Parser.parse(raw)
            model.assert_valid()
            return model
        raise InputError(f"Parsing error, get unknown version: {raw['version']}")

    @staticmethod
    def parse_from_file(fpath: Union[Path, str]) -> "DRepr":
        fpath = str(fpath)
        if fpath.endswith(".json"):
            with open(fpath, "rb") as f:
                return DRepr.parse(orjson.loads(f.read()))

        if fpath.endswith(".yml") or fpath.endswith(".yaml"):
            with open(fpath, "r") as f:
                return DRepr.parse(yaml.load(f))

        raise Exception(
            f"Does not supported this file: {fpath}. Only support json or yaml file"
        )

    @staticmethod
    def empty() -> "DRepr":
        return DRepr([], [], [], [], SemanticModel({}, {}, {}))

    @staticmethod
    def deserialize(raw: dict) -> "DRepr":
        resources = [Resource.deserialize(o) for o in raw["resources"]]
        preprocessing = [Preprocessing.deserialize(o) for o in raw["preprocessing"]]
        attrs = [Attr.deserialize(o) for o in raw["attrs"]]
        aligns = []

        for align in raw["aligns"]:
            if align["type"] == AlignmentType.Range.value:
                aligns.append(
                    RangeAlignment(
                        align["source"],
                        align["target"],
                        [
                            AlignedStep(step["source_idx"], step["target_idx"])
                            for step in align["aligned_steps"]
                        ],
                    )
                )
            elif align["type"] == AlignmentType.Value.value:
                aligns.append(ValueAlignment(align["source"], align["target"]))
            else:
                raise NotImplementedError()
        sm = SemanticModel.deserialize(raw["sm"])

        return DRepr(resources, preprocessing, attrs, aligns, sm)

    def serialize(self) -> dict:
        obj = asdict(self)
        # post-process the enum
        for res in obj["resources"]:
            res["type"] = res["type"].value
        for prepro in obj["preprocessing"]:
            prepro["type"] = prepro["type"].value
            for i, step in enumerate(prepro["value"]["path"]["steps"]):
                if isinstance(step, Enum):
                    prepro["value"]["path"]["steps"][i] = step.value

        for attr in obj["attrs"]:
            attr["sorted"] = attr["sorted"].value
            for i, step in enumerate(attr["path"]["steps"]):
                if isinstance(step, Enum):
                    attr["steps"][i] = step.value

            attr["value_type"] = attr["value_type"].value
        for node in obj["sm"]["nodes"].values():
            if node.get("data_type", None) is not None:
                node["data_type"] = node["data_type"].get_rel_uri()

        # adding a bit of meta-data about the alignment
        for align, raw_align in zip(self.aligns, obj["aligns"]):
            if isinstance(align, RangeAlignment):
                raw_align["type"] = AlignmentType.Range.value
            elif isinstance(align, ValueAlignment):
                raw_align["type"] = AlignmentType.Value.value
            else:
                raise NotImplementedError()

        # similarly, add meta-data about the nodes
        for node in obj["sm"]["nodes"].values():
            if isinstance(self.sm.nodes[node["node_id"]], ClassNode):
                node["type"] = "class_node"
            elif isinstance(self.sm.nodes[node["node_id"]], DataNode):
                node["type"] = "data_node"
            elif isinstance(self.sm.nodes[node["node_id"]], LiteralNode):
                node["type"] = "literal_node"
            else:
                raise NotImplementedError()

        return obj

    def assert_valid(self):
        """
        Perform a check to see if this D-REPR is valid. Raise AssertionError if this is not valid
        """
        # CHECK 1: all references (resource id, attribute ids) are valid
        resource_ids = {r.id for r in self.resources}

        attr_ids = {attr.id for attr in self.attrs}
        assert len(attr_ids) == len(self.attrs), "Duplicate attribute ids"
        for attr in self.attrs:
            assert (
                attr.resource_id in resource_ids
            ), f"Attribute {attr.resource_id} does not belong to any resources"

        for pref in self.preprocessing:
            if pref.value.output is not None:
                if pref.value.output.resource_id is not None:
                    assert (
                        pref.value.output.resource_id not in resource_ids
                    ), f"Preprocessing {pref} overwrite existence resource: {pref.value.output.resource_id}"
                if pref.value.output.attr is not None:
                    # preprocessing create new attribute
                    assert (
                        pref.value.output.attr not in attr_ids
                    ), f"Cannot overwrite existing attribute: {pref.value.output.attr}"
                    attr_ids.add(pref.value.output.attr)

        for align in self.aligns:
            if isinstance(align, AutoAlignment):
                if align.attrs is not None:
                    assert all(
                        attr_id in attr_ids for attr_id in align.attrs
                    ), f"The alignment {align} links to non-existence attributes"
            else:
                assert (
                    not isinstance(align, IdenticalAlign)
                    and align.source in attr_ids
                    and align.target in attr_ids
                ), f"The alignment {align} links to non-existence attributes"

        for node in self.sm.nodes.values():
            if isinstance(node, DataNode):
                assert node.attr_id in attr_ids, (
                    f"The semantic model has a link to "
                    f"a non-existence attribute: {node.attr_id}"
                )

        # CHECK 2: check class and predicates are valid
        for node in self.sm.nodes.values():
            if isinstance(node, ClassNode):
                if self.sm.is_rel_iri(node.label):
                    prefix = node.label.split(":", 1)[0]
                    assert prefix in self.sm.prefixes, (
                        f"Unknown prefix `{prefix}` of the "
                        f"ontology class {node.label}"
                    )
        for edge in self.sm.edges.values():
            if self.sm.is_rel_iri(edge.label):
                prefix = edge.label.split(":", 1)[0]
                assert prefix in self.sm.prefixes, (
                    f"Unknown prefix `{prefix}` of the "
                    f"ontology predicate {edge.label}"
                )
            assert edge.source_id in self.sm.nodes, edge
            assert edge.target_id in self.sm.nodes, edge

    def to_lang_format(
        self, simplify: bool = True, use_json_path: bool = False
    ) -> dict:
        return ReprV2Parser.dump(self, simplify, use_json_path)

    def to_lang_yml(self, simplify: bool = True, use_json_path: bool = False) -> str:
        model = self.to_lang_format(simplify, use_json_path)
        out = StringIO()
        yaml.dump(model, out)
        return out.getvalue()

    def _serde_engine_value(self, value: Any):
        """Serialize a python value to a json representation of the Value struct in the Rust engine"""
        if value is None:
            return {"t": "Null"}
        elif isinstance(value, bool):
            return {"t": "Bool", "c": value}
        elif isinstance(value, int):
            return {"t": "I64", "c": value}
        elif isinstance(value, float):
            return {"t": "F64", "c": value}
        elif isinstance(value, str):
            return {"t": "Str", "c": value}
        elif isinstance(value, list):
            return {"t": "Array", "c": [self._serde_engine_value(v) for v in value]}
        elif isinstance(value, (dict, OrderedDict)):
            return {
                "t": "Object",
                "c": {k: self._serde_engine_value(v) for k, v in value.items()},
            }
        else:
            raise InputError(
                f"Cannot serialize the value of type: {type(value)} to JSON"
            )

    def remove_resource(self, resource_id: str):
        self.resources = [r for r in self.resources if r.id != resource_id]
        for i in range(len(self.preprocessing) - 1, -1, -1):
            if self.preprocessing[i].value.resource_id == resource_id:
                self.preprocessing.pop(i)

        for i in range(len(self.attrs) - 1, -1, -1):
            if self.attrs[i].resource_id == resource_id:
                self.remove_attribute(self.attrs[i].id, idx=i)

    def get_resource_by_id(self, resource_id: str) -> Optional[Resource]:
        for r in self.resources:
            if r.id == resource_id:
                return r
        return None

    def add_resource(self, resource: Resource):
        assert self.get_resource_by_id(resource.id) is None
        self.resources.append(resource)

    def has_attr(self, attr_id: str) -> bool:
        return any(a.id == attr_id for a in self.attrs)

    def get_attr_index_by_id(self, attr_id: str) -> int:
        for i, a in enumerate(self.attrs):
            if a.id == attr_id:
                return i
        raise KeyError(f"Attribute with id {attr_id} does not exist")

    def get_attr_by_id(self, attr_id: str) -> Attr:
        for a in self.attrs:
            if a.id == attr_id:
                return a
        raise KeyError(f"Attribute with id {attr_id} does not exist")

    def remove_attribute(self, attr_id: str, idx: Optional[int] = None):
        if idx is None:
            idx = next(
                i for i in range(len(self.attrs), -1, -1) if self.attrs[i].id == attr_id
            )

        self.attrs.pop(idx)
        for i in range(len(self.aligns) - 1, -1, -1):
            align = self.aligns[i]
            if not isinstance(align, AutoAlignment):
                if align.source == attr_id or align.target == attr_id:
                    self.aligns.pop(i)

        for node in self.sm.nodes:
            if isinstance(node, DataNode) and node.attr_id == attr_id:
                self.sm.remove_node(node.node_id)

    def update_attribute(self, attr_id: str, new_attr: Attr):
        for i, attr in enumerate(self.attrs):
            if attr.id == attr_id:
                self.attrs[i] = new_attr

        for align in self.aligns:
            if isinstance(align, AutoAlignment):
                if align.attrs is not None:
                    align.attrs = [
                        new_attr.id if a == attr_id else a for a in align.attrs
                    ]
            else:
                if align.source == attr_id:
                    align.source = new_attr.id
                elif align.target == attr_id:
                    align.target = new_attr.id

        for node in self.sm.nodes:
            if isinstance(node, DataNode) and node.attr_id == attr_id:
                node.attr_id = new_attr.id



from __future__ import annotations

from codegen.models import AST, Program, Var, expr
from codegen.models.var import DeferredVar

from drepr.models.prelude import DRepr, OutputFormat
from drepr.program_generation.program_space import VarSpace


class Writer:
    def __init__(self, desc: DRepr, format: OutputFormat, program: Program):
        self.desc = desc
        self.format = format
        self.program = program

    def get_writer_clspath(self):
        if self.format == OutputFormat.TTL:
            return f"drepr.writers.turtle_writer.TurtleWriter"
        else:
            raise NotImplementedError()

    def create_writer(self, ast: AST):
        self.program.import_(self.get_writer_clspath(), True)
        writer_clsname = self.get_writer_clspath().rsplit(".", 1)[-1]
        ast.assign(
            DeferredVar(name="writer", key=VarSpace.writer()),
            expr.ExprFuncCall(
                expr.ExprIdent(writer_clsname),
                [expr.ExprConstant(self.desc.sm.prefixes)],
            ),
        )

    def has_written_record(self, ast: AST, subj: expr.Expr):
        return expr.ExprMethodCall(
            expr.ExprVar(
                self.program.get_var(
                    key=VarSpace.writer(),
                    at=ast.next_child_id(),
                )
            ),
            "has_written_record",
            [subj],
        )

    def begin_record(
        self,
        prog: AST,
        class_uri: expr.Expr,
        subj: expr.Expr,
        is_blank: expr.Expr,
        is_buffered: bool,
    ):
        """whether to bufferef the records because some properties are mandatory."""
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(
                        key=VarSpace.writer(),
                        at=prog.next_child_id(),
                    )
                ),
                "begin_record",
                [class_uri, subj, is_blank, expr.ExprConstant(is_buffered)],
            )
        )

    def end_record(self, prog: AST):
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "end_record",
                [],
            )
        )

    def abort_record(self, prog: AST):
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "abort_record",
                [],
            )
        )

    def is_record_empty(self, prog: AST):
        return expr.ExprMethodCall(
            expr.ExprVar(
                self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
            ),
            "is_record_empty",
            [],
        )

    def begin_partial_buffering_record(
        self,
        prog: AST,
    ):
        raise NotImplementedError()

    def write_data_property(
        self,
        prog: AST,
        predicate_id: expr.Expr,
        value: expr.Expr,
        dtype: expr.ExprConstant,
    ):
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "write_data_property",
                [predicate_id, value, dtype],
            )
        )

    def write_object_property(
        self,
        prog: AST,
        predicate_id: expr.Expr,
        object: expr.Expr,
        is_subject_blank: expr.Expr,
        is_object_blank: expr.Expr,
        is_new_subj: expr.Expr,
    ):
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "write_object_property",
                [predicate_id, object, is_subject_blank, is_object_blank, is_new_subj],
            )
        )

    def buffer_object_property(
        self,
        target_cls: str,
        predicate_id: str,
        object: str,
        is_object_blank: bool,
    ):
        raise NotImplementedError()

    def write_to_file(self, prog: AST, file_path: expr.Expr):
        prog.expr(
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "write_to_file",
                [file_path],
            )
        )

    def write_to_string(self, prog: AST, content: DeferredVar | Var):
        prog.assign(
            content,
            expr.ExprMethodCall(
                expr.ExprVar(
                    self.program.get_var(key=VarSpace.writer(), at=prog.next_child_id())
                ),
                "write_to_string",
                [],
            ),
        )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional

from graph.retworkx import RetworkXStrDiGraph

from drepr.models.prelude import (
    AlignedStep,
    Alignment,
    Attr,
    AttrId,
    AutoAlignment,
    Cardinality,
    DRepr,
    IdenticalAlign,
    IndexExpr,
    PreprocessResourceOutput,
    RangeAlignment,
    RangeExpr,
    SetIndexExpr,
    ValueAlignment,
)


@dataclass
class DReprModelAlignments:
    desc: DRepr
    aligns: dict[tuple[AttrId, AttrId], list[Alignment]]

    @staticmethod
    def create(desc: DRepr) -> DReprModelAlignments:
        """Infer all possible alignments from the model."""
        aligns = {(src.id, tgt.id): [] for src in desc.attrs for tgt in desc.attrs}

        for a in desc.aligns:
            if isinstance(a, RangeAlignment):
                aligns[(a.source, a.target)] = [a]
                aligns[(a.target, a.source)] = [a.swap()]
            elif isinstance(a, ValueAlignment):
                aligns[(a.source, a.target)] = [a]
                aligns[(a.target, a.source)] = [a.swap()]
            elif isinstance(a, AutoAlignment):
                if a.attrs is None:
                    attrs = desc.attrs
                else:
                    attrs = [desc.get_attr_by_id(attr) for attr in a.attrs]

                # now for each pair of attributes, we look at there steps and align them one by one
                for source in attrs:
                    for target in attrs:
                        if (
                            source.id == target.id
                            or len(aligns[source.id, target.id]) > 0
                        ):
                            continue

                        if source.resource_id != target.resource_id:
                            # check if source & target is derived from the same resource
                            source_resource = desc.get_resource_by_id(
                                source.resource_id
                            )
                            assert source_resource is not None
                            if not (
                                isinstance(source_resource, PreprocessResourceOutput)
                                and source_resource.get_preprocessing_original_resource_id()
                                == target.resource_id
                            ):
                                continue

                        new_align = DReprModelAlignments.auto_align(source, target)
                        if new_align is not None:
                            aligns[(source.id, target.id)] = [new_align]
                            aligns[(target.id, source.id)] = [new_align.swap()]
            elif isinstance(a, IdenticalAlign):
                raise Exception(
                    "Unreachable! Users should not provide identical alignment in the model"
                )

        for a in desc.attrs:
            aligns[(a.id, a.id)] = [IdenticalAlign(a.id, a.id)]

        inf = DReprModelAlignments(desc, aligns)
        inf.inference()
        return inf

    def get_alignments(self, source: AttrId, target: AttrId) -> list[Alignment]:
        return self.aligns[(source, target)]

    def inference(self):
        ig: dict[str, dict[str, int]] = {
            attr.id: {attr.id: 0} for attr in self.desc.attrs
        }
        for (source, target), aligns in self.aligns.items():
            assert len(aligns) <= 1, aligns
            if source == target or len(aligns) == 0:
                continue
            if isinstance(aligns[0], (RangeAlignment, ValueAlignment, IdenticalAlign)):
                ig[source][target] = sum(
                    self.score_alignment(align) for align in aligns
                )
            elif not isinstance(aligns[0], IdenticalAlign):
                print(aligns)
                raise Exception("Unreachable")

        ig_2nd = ig
        while True:
            ig_3rd = defaultdict(dict)
            for u0 in ig_2nd:
                # for updated neighbors of u0
                for u1 in ig_2nd[u0]:
                    for u2 in ig[u1]:
                        afuncs = self.infer_func(u0, u1, u2)
                        if afuncs is not None:
                            score = sum(self.score_alignment(align) for align in afuncs)
                            if u2 not in ig[u0] or ig[u0][u2] > score:
                                # we found a better alignment
                                self.aligns[u0, u2] = afuncs
                                ig_3rd[u0][u2] = score

            for u0 in ig_3rd:
                for u2, score in ig_3rd[u0].items():
                    ig[u0][u2] = score
            ig_2nd = ig_3rd
            if len(ig_2nd) == 0:
                break

    def infer_subject(self, attrs: list[AttrId]) -> list[AttrId]:
        """
        Find all attributes that can be the subject of these attributes

        The subject must has one to one mapping to all attrs and:
        (1) have no missing path
        (2) have missing path, but the missing path is the same for all attrs.

        (1) has higher priority than (2).
        """
        subjs = []

        # if one or these attributes has *-to-one, then that is the subjects
        for a in attrs:
            is_subj = True
            for ai in attrs:
                if len(self.aligns[(a, ai)]) == 0:
                    # no alignment
                    is_subj = False
                    break

                cardin = self.estimate_cardinality(self.aligns[(a, ai)])
                if cardin == Cardinality.ManyToMany:
                    is_subj = False
                    break
                elif cardin == Cardinality.OneToMany:
                    is_subj = False
                    break
            if is_subj:
                subjs.append(a)

        if len(subjs) == 0:
            attr_ids = set(attrs)

            # we have to try the external attributes
            for attr in self.desc.attrs:
                if attr.id in attr_ids:
                    continue

                is_candidate_subj = True
                covered_dims = set()

                for ai in attrs:
                    # we can only infer if there are any duplications if the alignments are dimensional
                    # if they are dimension alignment, then the optimization engine must compress them into
                    # just one alignment
                    if len(self.aligns[attr.id, ai]) != -1 or not isinstance(
                        self.aligns[attr.id, ai][0], RangeAlignment
                    ):
                        is_candidate_subj = False
                        break

                    align = self.aligns[attr.id, ai][0]
                    assert isinstance(align, RangeAlignment)
                    align_cardin = align.compute_cardinality(self.desc)
                    if align_cardin == Cardinality.ManyToMany:
                        is_candidate_subj = False
                        break
                    elif align_cardin == Cardinality.OneToMany:
                        is_candidate_subj = False
                        break
                    else:
                        for ad in align.aligned_steps:
                            covered_dims.add(ad.source_idx)

                if is_candidate_subj and covered_dims == set(
                    attr.path.get_nary_steps()
                ):
                    # the second condition detect if there is duplication
                    subjs.append(attr.id)

        if len(subjs) > 0:
            # now, we filter the missing path condition, if there is subject that have no missing path, we use them.
            if any(
                not self.desc.get_attr_by_id(subj).path.has_optional_steps()
                for subj in subjs
            ):
                return [
                    subj
                    for subj in subjs
                    if not self.desc.get_attr_by_id(subj).path.has_optional_steps()
                ]

            # check if we have subject with missing path, but the missing path is the same or less restricted for all attrs
            filtered_subjs = []

            for subj_id in subjs:
                subj = self.desc.get_attr_by_id(subj_id)

                # now we need to check if this subject has the same or less restricted missing path to all attrs.
                if all(
                    subj.path.has_same_or_less_optional_steps(
                        self.desc.get_attr_by_id(ai).path
                    )
                    for ai in attrs
                ):
                    filtered_subjs.append(subj_id)

            subjs = filtered_subjs

        return subjs

    def infer_func(
        self, xid: AttrId, yid: AttrId, zid: AttrId
    ) -> Optional[list[Alignment]]:
        """Infer an alignment function of xid and zid given alignments between (xid, yid) and (yid, zid)

        If there is only one way to join values of xid and zid, then the chain join will be the correct one
        """

        f = self.aligns[xid, yid]
        g = self.aligns[yid, zid]

        f_cardin = self.estimate_cardinality(f)
        g_cardin = self.estimate_cardinality(g)

        # filter the case where we cannot chain these alignments
        can_chain_alignments = (
            f_cardin == Cardinality.OneToOne or f_cardin == Cardinality.OneToMany
        ) or (g_cardin == Cardinality.OneToOne or g_cardin == Cardinality.ManyToOne)

        if not can_chain_alignments:
            return None

        # chain them together
        return self.optimize_chained_alignments(f + g)

    def estimate_cardinality(self, aligns: list[Alignment]) -> Cardinality:
        cardin = aligns[0].compute_cardinality(self.desc)
        # must always be > 0
        if len(aligns) <= 1 or cardin == Cardinality.ManyToMany:
            return cardin

        for i in range(1, len(aligns)):
            cardin_i = aligns[i].compute_cardinality(self.desc)

            if cardin_i == Cardinality.OneToOne:
                # do nothing, as this does not change the cardin
                pass
            elif cardin_i == Cardinality.OneToMany:
                if cardin == Cardinality.OneToOne:
                    cardin = Cardinality.OneToMany
                elif cardin == Cardinality.OneToMany:
                    cardin = Cardinality.OneToMany
                elif cardin == Cardinality.ManyToOne:
                    # we don't know whether it is going to be M2M, O2O, O2M, M2O so we do a conservative prediction
                    return Cardinality.ManyToMany
                else:
                    raise Exception("Unreachable")
            elif cardin_i == Cardinality.ManyToOne:
                if cardin == Cardinality.OneToOne:
                    cardin = Cardinality.ManyToOne
                elif cardin == Cardinality.OneToMany:
                    return Cardinality.ManyToMany
                elif cardin == Cardinality.ManyToOne:
                    cardin = Cardinality.ManyToOne
                else:
                    raise Exception("Unreachable")
            elif cardin_i == Cardinality.ManyToMany:
                return cardin_i

        return cardin

    def optimize_chained_alignments(self, aligns: list[Alignment]) -> list[Alignment]:
        if len(aligns) == 0:
            return []

        joins = []

        # rule 1: consecutive dimension alignments are combined together
        joins.append(aligns[0])

        for align in aligns[1:]:
            lastjoin = joins[-1]
            if isinstance(lastjoin, RangeAlignment) and isinstance(
                align, RangeAlignment
            ):
                # we merge them together
                a0 = lastjoin
                a1 = align
                a1map = {ad.source_idx: ad.target_idx for ad in a1.aligned_steps}

                joins[-1] = RangeAlignment(
                    source=a0.source,
                    target=a1.target,
                    aligned_steps=[
                        AlignedStep(
                            source_idx=ad.source_idx, target_idx=a1map[ad.target_idx]
                        )
                        for ad in a0.aligned_steps
                        if ad.target_idx in a1map
                    ],
                )
            else:
                joins.append(align)

        return joins

    @staticmethod
    def auto_align(
        source: Attr, target: Attr
    ) -> Optional[RangeAlignment | IdenticalAlign]:
        # TODO: fix auto-alignment
        source_range_steps = [
            i for i, step in enumerate(source.path.steps) if isinstance(step, RangeExpr)
        ]
        target_range_steps = [
            i for i, step in enumerate(target.path.steps) if isinstance(step, RangeExpr)
        ]

        if len(source_range_steps) == 0 or len(target_range_steps) == 0:
            # mapping between two single values is identical mapping
            is_source_single_val = all(
                isinstance(step, IndexExpr) for step in source.path.steps
            )
            is_target_single_val = all(
                isinstance(step, IndexExpr) for step in target.path.steps
            )
            if is_source_single_val and is_target_single_val:
                return IdenticalAlign(source.id, target.id)

            if is_source_single_val or is_target_single_val:
                return RangeAlignment(source.id, target.id, [])
            return None

        end = min(source_range_steps[-1], target_range_steps[-1]) + 1
        assert end > 0

        if source.path.steps[:end] == target.path.steps[:end]:
            # we can align all of them as long as all the steps are index
            if all(
                isinstance(step, (RangeExpr, IndexExpr, SetIndexExpr))
                for step in source.path.steps
            ) and all(
                isinstance(step, (RangeExpr, IndexExpr, SetIndexExpr))
                for step in target.path.steps
            ):
                return RangeAlignment(
                    source=source.id,
                    target=target.id,
                    aligned_steps=[
                        AlignedStep(source_idx=dim, target_idx=dim)
                        for dim in source_range_steps[
                            : min(len(target_range_steps), len(source_range_steps))
                        ]
                    ],
                )

        return None

    def score_alignment(self, align: Alignment) -> int:
        """Score an alignment function so we can justify if an alignment is new/better or not. Lower score is better."""
        if isinstance(align, RangeAlignment):
            card = align.compute_cardinality(self.desc)
            if card == Cardinality.OneToOne:
                return 1
            elif card == Cardinality.OneToMany or card == Cardinality.ManyToOne:
                return 2
            else:
                return 3
        if isinstance(align, ValueAlignment):
            card = align.compute_cardinality(self.desc)
            if card == Cardinality.OneToOne:
                return 4
            elif card == Cardinality.OneToMany or card == Cardinality.ManyToOne:
                return 5
            else:
                return 6
        assert isinstance(align, IdenticalAlign)
        return 0


@dataclass
class CustomedDfs:
    """
    THE FOLLOWING CODE IS ADAPTED FROM: https://docs.rs/petgraph/0.4.13/src/petgraph/visit/traversal.rs.html#38-43
    so that DFS also yield the parent node

    Visit nodes of a graph in a depth-first-search (DFS) emitting nodes in
    preorder (when they are first discovered).

    The traversal starts at a given node and only traverses nodes reachable
    from it.

    `Dfs` is not recursive.

    `Dfs` does not itself borrow the graph, and because of this you can run
    a traversal over a graph while still retaining mutable access to it, if you
    use it like the following example:

    ```
    use petgraph::Graph;
    use petgraph::visit::Dfs;

    let mut graph = Graph::<_,()>::new();
    let a = graph.add_node(0);

    let mut dfs = Dfs::new(&graph, a);
    while let Some(nx) = dfs.next(&graph) {
        // we can access `graph` mutably here still
        graph[nx] += 1;
    }

    assert_eq!(graph[a], 1);
    ```

    **Note:** The algorithm may not behave correctly if nodes are removed
    during iteration. It may not necessarily visit added nodes or edges.
    """

    stack: list[tuple[str, str]]
    discovered: VisitMap

    @staticmethod
    def empty() -> CustomedDfs:
        """Create a new **Dfs** using the graph's visitor map, and no stack."""
        return CustomedDfs([], VisitMap())

    @staticmethod
    def from_graph(graph: RetworkXStrDiGraph, start: str) -> CustomedDfs:
        """
        Create a new **Dfs**, using the graph's visitor map, and put **start**
        in the stack of nodes to visit.
        """
        dfs = CustomedDfs.empty()
        dfs.move_to(start)
        return dfs

    def move_to(self, start: str):
        """
        Keep the discovered map, but clear the visit stack and restart
        the dfs from a particular node.
        """
        self.discovered.visit(start)
        self.stack.clear()
        self.stack.append((start, start))

    def next(
        self, graph: RetworkXStrDiGraph, revisit: set[str]
    ) -> Optional[tuple[str, str]]:
        if len(self.stack) == 0:
            return None

        (parent_node, node) = self.stack.pop()
        for e in graph.out_edges(node):
            succ = e.target
            if self.discovered.visit(succ) or succ in revisit:
                self.stack.append((node, succ))
        return (parent_node, node)


@dataclass
class VisitMap:
    """A mapping for storing the visited status for NodeId `N`."""

    map: set[str] = field(default_factory=set)

    def visit(self, a: str) -> bool:
        """
        Mark `a` as visited.

        Return **true** if this is the first visit, false otherwise.
        """
        if a in self.map:
            return False
        self.map.add(a)
        return True

    def is_visited(self, a: str) -> bool:
        """Return whether `a` has been visited before."""
        return a in self.map



import copy
import re

from drepr.models.sm import (
    DREPR_URI,
    ClassNode,
    DataNode,
    DataType,
    Edge,
    LiteralNode,
    PredefinedDataType,
    SemanticModel,
)
from drepr.utils.misc import F, get_abs_iri
from drepr.utils.namespace_mixin import NamespaceMixin
from drepr.utils.validator import InputError, Validator


class SMParser:
    """
    SM has the following schema

    ```
    semantic_model:
      <class_id>:
        properties:
            - [<predicate>, <attr_id>, (<data_type>, (<is_required=false>)?)?]
        links:
            - [<predicate>, <class_id>, (<is_required=false>)?]
        static_properties:
            - [<predicate>, <value>, (<data_type>)?]
        inverse_static_properties:
            - [<predicate>, <uri>]
      prefixes:
        <prefix>: <uri>
    ```
    """

    CLS_KEYS = {
        "properties",
        "subject",
        "links",
        "static_properties",
        "inverse_static_properties",
    }

    REG_SM_CLASS = re.compile(r"^((.+):[a-zA-Z0-9_]+)$")
    REG_SM_DNODE = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_LNODE = re.compile(
        r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$"
    )
    REG_SM_REL = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!--).)+:\d+)$")

    @classmethod
    def parse(cls, sm: dict) -> SemanticModel:
        nodes = {}
        edges = {}
        # shallow copy
        sm = copy.copy(sm)
        prefixes = sm.pop("prefixes", {})

        trace0 = f"Parsing `prefixes` of the semantic model"
        Validator.must_be_dict(prefixes, trace0)
        for prefix, uri in prefixes.items():
            Validator.must_be_str(uri, f"{trace0}\nParse prefix {prefix}")

        for prefix, uri in SemanticModel.get_default_prefixes().items():
            if prefix not in prefixes:
                prefixes[prefix] = uri
            elif prefixes[prefix] != uri:
                raise InputError(
                    f"{trace0}\nERROR: Prefix `{prefix}` is conflicting with predefined value `{uri}`."
                )

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"
            Validator.must_be_dict(class_conf, trace0)
            Validator.must_be_subset(
                cls.CLS_KEYS, class_conf.keys(), "keys of ontology class", trace0
            )

            try:
                m = cls.REG_SM_CLASS.match(class_id)
                assert m is not None
                class_name = m.group(2)
            except Exception as e:
                raise InputError(
                    f"{trace0}\nERROR: invalid class_id `{class_id}`. Expect to be <string>:<alphanumeric & _>"
                )

            nodes[class_id] = ClassNode(class_id, class_name)

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"

            for i, prop in enumerate(class_conf.get("properties", [])):
                trace1 = f"{trace0}\nParsing property {i}: {prop}"
                if len(prop) == 2:
                    predicate, attr_id = prop
                    data_type = None
                    is_required = False
                elif len(prop) == 3:
                    predicate, attr_id, data_type = prop
                    if isinstance(data_type, bool) or data_type.lower() in {
                        "true",
                        "false",
                    }:
                        is_required = (
                            data_type
                            if isinstance(data_type, bool)
                            else data_type == "true"
                        )
                        data_type = None
                    else:
                        is_required = False
                elif len(prop) == 4:
                    predicate, attr_id, data_type, is_required = prop
                else:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the property to be an array of two "
                        f"three or four items (<predicate>, <attribute_id>[, data_type='auto'][, is_required=false])"
                    )

                if data_type is not None:
                    try:
                        data_type = DataType(data_type, prefixes)
                    except Exception as e:
                        raise InputError(f"{trace1}\nERROR: {str(e)}")

                node = DataNode(
                    node_id=f"dnode:{attr_id}", attr_id=attr_id, data_type=data_type
                )
                nodes[node.node_id] = node
                edges[len(edges)] = Edge(
                    len(edges),
                    class_id,
                    node.node_id,
                    predicate,
                    is_required=is_required,
                )

            for i, link_conf in enumerate(class_conf.get("links", [])):
                trace1 = f"{trace0}\nParsing link {i}: {link_conf}"
                if not (2 <= len(link_conf) <= 3):
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the link to be an array of two or three"
                        f"items (<predicate>, <class_id>, <is_required=false>)"
                    )

                if len(link_conf) == 2:
                    predicate, object_class_id = link_conf
                    is_required = False
                else:
                    predicate, object_class_id, is_required = link_conf

                edges[len(edges)] = Edge(
                    len(edges),
                    class_id,
                    object_class_id,
                    predicate,
                    is_required=is_required,
                )

            for i, prop in enumerate(class_conf.get("static_properties", [])):
                trace1 = f"{trace0}\nParsing static properties {i}: {prop}"
                if len(prop) == 2:
                    predicate, value = prop
                    data_type = None
                elif len(prop) == 3:
                    predicate, value, data_type = prop
                else:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the property to be an array of two "
                        f"or three items (<predicate>, <value>, [data_type])"
                    )

                if data_type is not None:
                    try:
                        data_type = DataType(data_type, prefixes)
                    except Exception as e:
                        raise InputError(f"{trace1}\nERROR: {str(e)}")

                # normalize value's type (e.g., ruamel.yaml read float into ScalarFloat)
                if isinstance(value, str):
                    value = str(value)
                elif isinstance(value, int):
                    value = int(value)
                elif isinstance(value, float):
                    value = float(value)

                if data_type == DREPR_URI:
                    assert isinstance(value, str)
                    # test if the value is relative uri, if so, convert it to absolute uri
                    if NamespaceMixin.is_rel_iri(value):
                        value = get_abs_iri(prefixes, value)

                node = LiteralNode(
                    node_id=f"lnode:{len(nodes)}", value=value, data_type=data_type
                )
                nodes[node.node_id] = node
                edges[len(edges)] = Edge(len(edges), class_id, node.node_id, predicate)

            for i, prop in enumerate(class_conf.get("inverse_static_properties", [])):
                trace1 = f"{trace0}\nParsing inverse static properties {i}: {prop}"
                if len(prop) == 2:
                    predicate, value = prop
                    data_type = PredefinedDataType.xsd_anyURI.value
                else:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the property to be an array of two "
                        f"items (<predicate>, <uri>)"
                    )

                if not isinstance(value, str):
                    raise InputError("Value of inverse static property must be an uri")
                value = str(value)
                node = LiteralNode(
                    node_id=f"lnode:{len(nodes)}", value=value, data_type=data_type
                )
                nodes[node.node_id] = node
                edges[len(edges)] = Edge(len(edges), node.node_id, class_id, predicate)

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"
            if "subject" in class_conf:
                trace1 = f"{trace0}\nParsing subject"
                attr_id = class_conf["subject"]
                nodes[class_id].subject = attr_id

        return SemanticModel(nodes, edges, prefixes)



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Callable

from codegen.models import AST, PredefinedFn, Program, expr, stmt
from codegen.models.var import DeferredVar
from drepr.models.prelude import (
    DRepr,
    IndexExpr,
    OutputFormat,
    PreprocessResourceOutput,
)
from drepr.planning.class_map_plan import (
    BlankObject,
    BlankSubject,
    ClassesMapExecutionPlan,
    ClassMapPlan,
    DataProp,
    ExternalIDSubject,
    IDObject,
    InternalIDSubject,
    LiteralProp,
    ObjectProp,
    SingletonObject,
    SingletonSubject,
)
from drepr.program_generation.alignment_fn import AlignmentFn, PathAccessor
from drepr.program_generation.predefined_fn import DReprPredefinedFn
from drepr.program_generation.preprocessing import GenPreprocessing
from drepr.program_generation.program_space import VarSpace
from drepr.program_generation.writers import Writer
from drepr.utils.misc import assert_true


@dataclass
class FileOutput:
    fpath: Path
    format: OutputFormat


@dataclass
class MemoryOutput:
    format: OutputFormat


Output = FileOutput | MemoryOutput


def gen_program(
    desc: DRepr, exec_plan: ClassesMapExecutionPlan, output: Output, debuginfo: bool
) -> AST:
    """Generate a program to convert the given D-REPR to a target format"""
    program = Program()
    writer = Writer(desc, output.format, program)

    func_args = [
        DeferredVar(
            name="resource" if len(desc.resources) == 1 else f"resource_{res.id}",
            key=VarSpace.resource(res.id),
        )
        for res in desc.resources
        if not isinstance(res, PreprocessResourceOutput)
    ]
    if isinstance(output, FileOutput):
        output_file = DeferredVar(name="output_file", key=VarSpace.output_file())
        func_args.append(output_file)
    else:
        output_file = None

    program.root.linebreak()
    main_fn = program.root.func("main", func_args)

    for resource in desc.resources:
        if isinstance(resource, PreprocessResourceOutput):
            continue
        var = DeferredVar(
            name=(
                "resource_data"
                if len(desc.resources) == 1
                else f"resource_data_{resource.id}"
            ),
            key=VarSpace.resource_data(resource.id),
        )
        main_fn.assign(
            var,
            DReprPredefinedFn.read_source(
                program,
                resource.type,
                program.get_var(
                    key=VarSpace.resource(resource.id), at=main_fn.next_child_id()
                ),
            ),
        )

    # define missing values of attributes
    main_fn.linebreak()
    for attr in desc.attrs:
        if len(attr.missing_values) > 0:
            main_fn.assign(
                DeferredVar(
                    name=f"{attr.id}_missing_values",
                    key=VarSpace.attr_missing_values(attr.id),
                ),
                expr.ExprConstant(set(attr.missing_values)),
            )

    # create transformation
    GenPreprocessing(program, desc, main_fn).generate()

    # create a writer
    writer.create_writer(main_fn)

    # for each class node, we generate a plan for each of them.
    for classplan in exec_plan.class_map_plans:
        main_fn.linebreak()
        main_fn.comment(f"Transform records of class {classplan.class_id}")
        # generate the code to execute the plan
        gen_classplan_executor(
            program, main_fn.block(), writer, desc, classplan, debuginfo
        )

    main_fn.linebreak()
    # we write the output to the file
    if isinstance(output, FileOutput):
        assert output_file is not None
        writer.write_to_file(main_fn, expr.ExprVar(output_file.get_var()))
    else:
        content = DeferredVar(name="output")
        writer.write_to_string(main_fn, content)
        main_fn.return_(expr.ExprVar(content.get_var()))

    invok_main = expr.ExprFuncCall(
        expr.ExprIdent("main"), [expr.ExprIdent("*sys.argv[1:]")]
    )

    program.root.linebreak()
    program.root.if_(
        expr.ExprEqual(expr.ExprIdent("__name__"), expr.ExprConstant("__main__"))
    )(
        stmt.ImportStatement("sys", False),
        stmt.LineBreak(),
        stmt.SingleExprStatement(
            expr.ExprFuncCall(expr.ExprIdent("print"), [invok_main])
            if isinstance(output, MemoryOutput)
            else invok_main
        ),
    )
    return program.root


def gen_classplan_executor(
    program: Program,
    parent_ast: AST,
    writer: Writer,
    desc: DRepr,
    classplan: ClassMapPlan,
    debuginfo: bool,
):
    """Generate the code to execute the given class plan.
    Below is the pseudo code:

    1. Iterate over the subject values
        1. If the subject is uri and it has missing values, if the uri is missing, we skip this
           record
        2. Begin record
        3. Iterate over target property & value
            1. If not target.can_have_missing_values:
                1. Iterate over objprop values:
                    1. Write property
            2. Else:
                1. If target edge is optional:
                    iterate over objprop values:
                        if objprop value is not missing:
                            write  property
                else:
                    (1) ----
                    has_record = False
                    iterate over objprop values:
                        if objprop value is not missing:
                            has_record = True
                            write property
                    if not has_record:
                        abort the record
                    ---- (2)
        4. End the record -- if the subject is blank node,
            and we do not write any data, we abort, otherwise, we commit
    """
    class_uri = expr.ExprConstant(
        desc.sm.get_abs_iri(desc.sm.get_class_node(classplan.class_id).label)
    )
    get_subj_val: Callable[[AST], expr.Expr]
    classplan_subject = classplan.subject
    if isinstance(classplan_subject, SingletonSubject):
        ast = parent_ast
        is_subj_blank = classplan_subject.is_blank
        is_buffered = False
        can_class_missing = False
        get_subj_val = lambda ast: get_subj_val_for_static_class(classplan.class_id)
    else:
        ast = PathAccessor(program).iterate_elements(
            parent_ast,
            classplan_subject.attr,
            None,
            None,
            validate_path=debuginfo,
            on_missing_key=(
                lambda tree: (
                    PathAccessor.skip_on_missing_key(parent_ast, tree)
                    if classplan_subject.attr.path.has_optional_steps()
                    else None
                )
            ),
        )
        is_subj_blank = isinstance(classplan_subject, BlankSubject)
        can_class_missing = (
            any(
                not dprop.is_optional and dprop.can_target_missing
                for dprop in classplan.data_props
            )
            or any(
                not oprop.is_optional and oprop.can_target_missing
                for oprop in classplan.object_props
            )
            or any(
                not oprop.is_optional and oprop.can_target_missing
                for oprop in classplan.buffered_object_props
            )
        )
        is_buffered = can_class_missing

        if isinstance(classplan_subject, (InternalIDSubject, ExternalIDSubject)):
            get_subj_val = lambda ast: expr.ExprVar(
                program.get_var(
                    key=VarSpace.attr_value_dim(
                        classplan_subject.attr.resource_id,
                        classplan_subject.attr.id,
                        len(classplan_subject.attr.path.steps) - 1,
                    ),
                    at=ast.next_child_id(),
                )
            )
        else:
            assert isinstance(classplan_subject, BlankSubject)
            if classplan_subject.use_attr_value:
                get_subj_val = lambda ast: PredefinedFn.tuple(
                    [
                        expr.ExprConstant(classplan.class_id),
                        expr.ExprConstant(
                            desc.get_attr_index_by_id(classplan_subject.attr.id)
                        ),
                        expr.ExprVar(
                            program.get_var(
                                key=VarSpace.attr_value_dim(
                                    classplan_subject.attr.resource_id,
                                    classplan_subject.attr.id,
                                    len(classplan_subject.attr.path.steps) - 1,
                                ),
                                at=ast.next_child_id(),
                            )
                        ),
                    ]
                )
            else:
                # if we don't use attr value, the subj_val is the entire index that leads to the last value
                get_subj_val = lambda ast: (
                    PredefinedFn.tuple(
                        [
                            expr.ExprConstant(classplan.class_id),
                            expr.ExprConstant(
                                desc.get_attr_index_by_id(classplan_subject.attr.id)
                            ),
                        ]
                        + [
                            expr.ExprVar(
                                program.get_var(
                                    key=VarSpace.attr_index_dim(
                                        classplan_subject.attr.resource_id,
                                        classplan_subject.attr.id,
                                        dim,
                                    ),
                                    at=ast.next_child_id(),
                                )
                            )
                            for dim, step in enumerate(
                                classplan_subject.attr.path.steps
                            )
                            if not isinstance(step, IndexExpr)
                        ]
                    )
                )

        if (
            isinstance(classplan_subject, (InternalIDSubject, ExternalIDSubject))
            and len(classplan_subject.attr.missing_values) > 0
        ) or (
            isinstance(classplan_subject, BlankSubject)
            and classplan_subject.use_attr_value
            and len(classplan_subject.attr.missing_values) > 0
        ):
            # we know immediately that it's missing if the subject value is missing

            if ast.id == parent_ast.id:
                # same ast because of a single value, we can't use continue
                # so we wrap it with if -- if not missing, continue to generate the instance
                ast = ast.if_(
                    expr.ExprNegation(
                        PredefinedFn.set_contains(
                            expr.ExprVar(
                                program.get_var(
                                    key=VarSpace.attr_missing_values(
                                        classplan_subject.attr.id
                                    ),
                                    at=ast.next_child_id(),
                                )
                            ),
                            get_subj_val(ast),
                        )
                    )
                )
            else:
                ast.if_(
                    PredefinedFn.set_contains(
                        expr.ExprVar(
                            program.get_var(
                                key=VarSpace.attr_missing_values(
                                    classplan_subject.attr.id
                                ),
                                at=ast.next_child_id(),
                            )
                        ),
                        get_subj_val(ast),
                    )
                )(stmt.ContinueStatement())

    writer.begin_record(
        ast,
        class_uri,
        get_subj_val(ast),
        expr.ExprConstant(is_subj_blank),
        is_buffered,
    )

    for dataprop in classplan.data_props:
        ast.linebreak()
        ast.comment(f"Retrieve value of data property: {dataprop.attr.id}")

        gen_classprop_body(
            program,
            desc,
            parent_ast,
            ast.block(),
            writer,
            is_buffered,
            is_subj_blank,
            dataprop,
            debuginfo,
        )

    for objprop in classplan.object_props:
        ast.linebreak()
        if isinstance(objprop, SingletonObject):
            ast.comment(
                f"Link object property to a singleton object: {objprop.target_class_id}"
            )
        else:
            ast.comment(f"Retrieve value of object property: {objprop.attr.id}")

        gen_classprop_body(
            program,
            desc,
            parent_ast,
            ast.block(),
            writer,
            is_buffered,
            is_subj_blank,
            objprop,
            debuginfo,
        )

    if len(classplan.literal_props) > 0:
        ast.linebreak()
        ast.comment("Set static properties")

        for litprop in classplan.literal_props:
            gen_classprop_body(
                program,
                desc,
                parent_ast,
                ast.block(),
                writer,
                is_buffered,
                is_subj_blank,
                litprop,
                debuginfo,
            )

    assert len(classplan.buffered_object_props) == 0, "Not implemented yet"

    # we can end the record even if we abort it before. the end record code should handle this.
    ast.linebreak()

    if isinstance(classplan_subject, BlankSubject) and can_class_missing:
        ast.if_(writer.is_record_empty(ast))(lambda ast00: writer.abort_record(ast00))
        ast.else_()(lambda ast00: writer.end_record(ast00))
    else:
        writer.end_record(ast)

    return ast


def gen_classprop_body(
    program: Program,
    desc: DRepr,
    parent_ast: AST,
    ast: AST,
    writer: Writer,
    is_buffered: bool,
    is_subj_blank: bool,
    classprop: DataProp | ObjectProp | LiteralProp,
    debuginfo: bool,
):
    """
    Args:
        parent_ast: the parent AST that above iterating subject values -- this is good to detect continue statement is okay to skip to the next subject/record
    """
    iter_final_list = False
    get_prop_val: Callable[[AST], expr.Expr]
    if isinstance(classprop, (DataProp, IDObject)):
        attr = classprop.attr
        if isinstance(classprop, DataProp) and classprop.attr.value_type.is_list():
            # for a list, we need to iterate over the list.
            get_prop_val = lambda ast: expr.ExprVar(
                program.get_var(
                    key=VarSpace.attr_value_dim(
                        attr.resource_id,
                        attr.id,
                        len(
                            attr.path.steps
                        ),  # not -1 because the last dimension is now a list
                    ),
                    at=ast.next_child_id(),
                )
            )
            iter_final_list = True
        else:
            get_prop_val = lambda ast: expr.ExprVar(
                program.get_var(
                    key=VarSpace.attr_value_dim(
                        attr.resource_id,
                        attr.id,
                        len(attr.path.steps) - 1,
                    ),
                    at=ast.next_child_id(),
                )
            )
    elif isinstance(classprop, BlankObject):
        attr = classprop.attr
        if classprop.use_attr_value:
            get_prop_val = lambda ast: PredefinedFn.tuple(
                [
                    expr.ExprConstant(classprop.object_id),
                    expr.ExprConstant(desc.get_attr_index_by_id(attr.id)),
                    expr.ExprVar(
                        program.get_var(
                            key=VarSpace.attr_value_dim(
                                attr.resource_id,
                                attr.id,
                                len(attr.path.steps) - 1,
                            ),
                            at=ast.next_child_id(),
                        )
                    ),
                ]
            )
        else:
            get_prop_val = lambda ast: (
                PredefinedFn.tuple(
                    [
                        expr.ExprConstant(classprop.object_id),
                        expr.ExprConstant(desc.get_attr_index_by_id(classprop.attr.id)),
                    ]
                    + [
                        expr.ExprVar(
                            program.get_var(
                                key=VarSpace.attr_index_dim(
                                    classprop.attr.resource_id,
                                    classprop.attr.id,
                                    dim,
                                ),
                                at=ast.next_child_id(),
                            )
                        )
                        for dim, step in enumerate(classprop.attr.path.steps)
                        if not isinstance(step, IndexExpr)
                    ]
                )
            )
    elif isinstance(classprop, LiteralProp):
        get_prop_val = lambda ast: expr.ExprConstant(classprop.value)
    else:
        assert isinstance(classprop, SingletonObject)
        get_prop_val = lambda ast: get_subj_val_for_static_class(
            classprop.target_class_id
        )

    is_prop_val_not_missing: Callable[[AST], expr.Expr]
    if isinstance(classprop, DataProp):
        if len(attr.missing_values) == 0:
            # leverage the fact that if True will be optimized away
            is_prop_val_not_missing = lambda ast: expr.ExprConstant(True)
        else:
            is_prop_val_not_missing = lambda ast: expr.ExprNegation(
                PredefinedFn.set_contains(
                    expr.ExprVar(
                        program.get_var(
                            key=VarSpace.attr_missing_values(attr.id),
                            at=ast.next_child_id(),
                        )
                    ),
                    get_prop_val(ast),
                ),
            )
        write_fn = partial(
            writer.write_data_property, dtype=expr.ExprConstant(classprop.datatype)
        )
    elif isinstance(classprop, ObjectProp):
        is_prop_val_not_missing = lambda ast: writer.has_written_record(
            ast,
            get_prop_val(ast),
        )
        write_fn = partial(
            writer.write_object_property,
            is_subject_blank=expr.ExprConstant(is_subj_blank),
            is_object_blank=expr.ExprConstant(classprop.is_object_blank()),
            is_new_subj=expr.ExprConstant(False),
        )
    else:
        assert isinstance(classprop, LiteralProp)
        is_prop_val_not_missing = lambda ast: expr.ExprConstant(True)
        write_fn = partial(
            writer.write_data_property, dtype=expr.ExprConstant(classprop.datatype)
        )

    if isinstance(classprop, (LiteralProp, SingletonObject)):
        write_fn(ast, expr.ExprConstant(classprop.predicate), get_prop_val(ast))
    else:
        if not classprop.can_target_missing:
            AlignmentFn(desc, program).align(
                ast, classprop.alignments, debuginfo, None, iter_final_list
            )(
                lambda ast_l0: write_fn(
                    ast_l0,
                    expr.ExprConstant(classprop.predicate),
                    get_prop_val(ast_l0),
                )
            )
        else:
            if classprop.is_optional:
                AlignmentFn(desc, program).align(
                    ast,
                    classprop.alignments,
                    debuginfo,
                    # if the value is missing, we just ignore it.
                    on_missing_key=lambda astxx: astxx(stmt.NoStatement()),
                    iter_final_list=iter_final_list,
                )(
                    lambda ast00: ast00.if_(is_prop_val_not_missing(ast00))(
                        lambda ast01: write_fn(
                            ast01,
                            expr.ExprConstant(classprop.predicate),
                            get_prop_val(ast01),
                        )
                    )
                )
            else:
                if classprop.alignments_cardinality.is_star_to_many():
                    has_dataprop_val = DeferredVar(
                        name=f"{attr.id}_has_value_d{len(attr.path.steps) - 1}",
                        key=VarSpace.has_attr_value_dim(
                            attr.resource_id,
                            attr.id,
                            len(attr.path.steps) - 1,
                        ),
                    )
                    ast.assign(has_dataprop_val, expr.ExprConstant(False))
                    has_dataprop_val = has_dataprop_val.get_var()

                    AlignmentFn(desc, program).align(
                        ast,
                        classprop.alignments,
                        debuginfo,
                        lambda astxx: astxx(stmt.NoStatement()),
                        iter_final_list,
                    )(
                        lambda ast00: ast00.if_(is_prop_val_not_missing(ast00))(
                            lambda ast01: ast01.assign(
                                has_dataprop_val, expr.ExprConstant(True)
                            ),
                            lambda ast02: write_fn(
                                ast02,
                                expr.ExprConstant(classprop.predicate),
                                get_prop_val(ast02),
                            ),
                        )
                    )
                    ast.if_(expr.ExprNegation(expr.ExprVar(has_dataprop_val)))(
                        lambda ast00: (
                            assert_true(
                                is_buffered,
                                "We should only abort record if we are buffering",
                            )
                            and writer.abort_record(ast00)
                        ),
                        (
                            stmt.ContinueStatement()
                            if parent_ast.has_statement_between_ast(
                                stmt.ForLoopStatement, ast.id
                            )
                            else stmt.NoStatement()
                        ),
                    )
                else:

                    def on_missing_key(tree: AST):
                        assert_true(
                            is_buffered,
                            "We should only abort record if we are buffering",
                        )
                        writer.abort_record(tree)
                        if parent_ast.has_statement_between_ast(
                            stmt.ForLoopStatement, tree.id
                        ):
                            tree(stmt.ContinueStatement())
                        else:
                            # same ast because of a single value, we can't use continue
                            # however, we use pass as it's a single-level if/else -- the else part
                            # will handle the instance generation if there is no missing value.
                            tree(stmt.NoStatement())

                    AlignmentFn(desc, program).align(
                        ast,
                        classprop.alignments,
                        debuginfo,
                        # on_missing_key=lambda astxx: assert_true(
                        #     is_buffered,
                        #     "We should only abort record if we are buffering",
                        # )
                        # and writer.abort_record(astxx),
                        on_missing_key=on_missing_key,
                        iter_final_list=iter_final_list,
                    )(
                        lambda ast00: ast00.if_(is_prop_val_not_missing(ast00))(
                            lambda ast01: write_fn(
                                ast01,
                                expr.ExprConstant(classprop.predicate),
                                get_prop_val(ast01),
                            ),
                        ),
                        lambda ast10: ast10.else_()(
                            lambda ast11: (
                                assert_true(
                                    is_buffered,
                                    "We should only abort record if we are buffering",
                                )
                                and writer.abort_record(ast11)
                            ),
                            (
                                stmt.ContinueStatement()
                                if parent_ast.has_statement_between_ast(
                                    stmt.ForLoopStatement, ast10.id
                                )
                                else stmt.NoStatement()
                            ),
                        ),
                    )


def get_subj_val_for_static_class(class_id):
    return PredefinedFn.tuple(
        [expr.ExprConstant("static-8172a"), expr.ExprConstant(class_id)]
    )



from typing import List

from drepr.models.resource import *
from drepr.utils.validator import InputError, Validator


class ResourceParser:
    """
    `resources` has two possible schemas

    1. Shorthand when you have only one resource (`resource_id` is `default`): `resources: <resource_type>`
    2. When you have multiple resources:
        ```
        resources:
            <resource_id>: <resource_conf>
            # .. other resources ..
        ```

        The `<resource_conf>` can either be:
        a. `<resource_type>`, when other resource properties are all options
        b. a dictionary as follows, when some properties require to define explicitly
            ```
                <type>: <resource_type>
                # .. other attributes ..
            ```
    """

    DEFAULT_RESOURCE_ID = "default"
    RESOURCE_TYPES = {rtype.value for rtype in ResourceType}

    @classmethod
    def parse(cls, resources: Union[str, dict]) -> List[Resource]:
        if isinstance(resources, str):
            return cls._parse_schema1(resources)

        if isinstance(resources, dict):
            return cls._parse_schema2(resources)

        raise InputError(
            f"Invalid type for `resources`, expect either a string or a dictionary. Get {type(resources)} instead"
        )

    @classmethod
    def _parse_schema1(cls, resource_type: str) -> List[Resource]:
        Validator.must_in(
            resource_type,
            cls.RESOURCE_TYPES,
            error_msg="Invalid resource type for `resources`",
        )
        resource_type = ResourceType(resource_type)
        if resource_type == ResourceType.CSV:
            resource_prop = CSVProp()
        else:
            resource_prop = None

        return [Resource(cls.DEFAULT_RESOURCE_ID, resource_type, resource_prop)]

    @classmethod
    def _parse_schema2(cls, resources: dict) -> List[Resource]:
        result = []
        for resource_id, conf in resources.items():
            trace = f"Parsing resource {resource_id}."

            if isinstance(conf, str):
                resource = cls._parse_schema1(conf)[0]
                resource.id = resource_id
            elif isinstance(conf, dict):
                Validator.must_have(conf, "type", trace)
                Validator.must_in(
                    conf["type"],
                    cls.RESOURCE_TYPES,
                    f"{trace}\n\tParsing resource type",
                )
                resource_type = ResourceType(conf["type"])

                if resource_type == ResourceType.CSV:
                    if "delimiter" in conf:
                        if len(conf["delimiter"]) != 1:
                            raise InputError(
                                f"{trace}.\nERROR: Expect one character delimiter "
                                f"for CSV resource. Get `{conf['delimiter']}` instead"
                            )

                        resource_prop = CSVProp(conf["delimiter"])
                    else:
                        resource_prop = CSVProp()
                else:
                    resource_prop = None

                resource = Resource(resource_id, resource_type, resource_prop)
            else:
                raise InputError(
                    f"{trace}.\nERROR: The configuration of a resource can either be string "
                    f"or dictionary. Get {type(conf)} instead"
                )

            result.append(resource)
        return result



from dataclasses import dataclass, field
from email.policy import default
from typing import List

from drepr.models.attr import Attr, Sorted, ValueType
from drepr.models.parsers.interface import PathParser
from drepr.models.resource import Resource
from drepr.utils.validator import InputError, Validator


@dataclass
class ParsedAttrs:
    attrs: list[Attr] = field(default_factory=list)
    id2attr: dict[str, Attr] = field(default_factory=dict)
    attrs_created_in_preprocessing: set[str] = field(default_factory=set)

    def add(self, attr: Attr):
        if attr.id in self.attrs:
            raise InputError(f"Duplicated attribute id: {attr.id}")

        self.attrs.append(attr)
        self.id2attr[attr.id] = attr

    def add_preprocessing_attr(self, attr_id: str):
        self.attrs_created_in_preprocessing.add(attr_id)

    def has_been_reference_before(self, attr_id: str):
        return attr_id in self.id2attr or attr_id in self.attrs_created_in_preprocessing

    def __contains__(self, id: str):
        return id in self.id2attr

    def __getitem__(self, id: str):
        return self.id2attr[id]


class AttrParser:
    """
    Attributes have two possible schemas
    1. When an attribute has only one path
        ```
        <attr_id>: <attr_path>
        # .. other attributes ..
        ```
    2.
        ```
        <attribute_id>:
            [resource_id]: <resource_id> (default is "default")
            path: <attr_path>
            [unique]: true|false (default is false)
            [sorted]: none|ascending|descending (default is none)
            [value_type]: unspecified|int|float|str|list[int]|list[str]|list[float] (default is unspecified)
            [missing_values]: [<value0>, <value1>, ...]
        ```
    """

    SORTED_VALUES = {x.value for x in Sorted}
    VALUE_TYPE_VALUES = {x.value for x in ValueType}
    CLS_KEYS = {
        "resource_id",
        "path",
        "unique",
        "sorted",
        "value_type",
        "missing_values",
    }

    def __init__(self, path_parser: PathParser):
        self.path_parser = path_parser

    def parse(
        self,
        default_resource_id: str,
        resources: List[Resource],
        parsed_attrs: ParsedAttrs,
        def_attrs: dict,
    ):
        Validator.must_be_dict(def_attrs, "Parsing attributes")
        for attr_id, attr_conf in def_attrs.items():
            trace = f"Parsing attribute: {attr_id}"

            if attr_id in parsed_attrs:
                raise InputError(f"{trace}\nERROR: Duplicated attribute id: {attr_id}")

            if isinstance(attr_conf, (str, list)):
                attr_path = self.path_parser.parse(
                    self.path_parser.get_resource(
                        resources, default_resource_id, trace
                    ),
                    attr_conf,
                    trace,
                )
                attr = Attr(attr_id, default_resource_id, attr_path, [])
            elif isinstance(attr_conf, dict):
                attr = self.parse_expanded_def(
                    default_resource_id, resources, attr_id, attr_conf, trace
                )
            else:
                raise InputError(
                    f"{trace}\nERROR: The configuration of an attribute can either be string, list, "
                    f"or dictionary. Get {type(attr_conf)} instead"
                )

            parsed_attrs.add(attr)

    def parse_expanded_def(
        self,
        default_resource_id: str,
        resources: List[Resource],
        attr_id: str,
        attr_conf: dict,
        parse_trace: str,
    ) -> Attr:
        resource_id = attr_conf.get("resource_id", default_resource_id)
        Validator.must_be_str(resource_id, f"{parse_trace}\nParsing `resource_id`")

        Validator.must_have(attr_conf, "path", parse_trace)
        path = self.path_parser.parse(
            self.path_parser.get_resource(resources, resource_id, parse_trace),
            attr_conf["path"],
            f"{parse_trace}\nParsing path of the attribute",
        )

        if "unique" in attr_conf and not isinstance(attr_conf["unique"], bool):
            raise InputError(
                f"{parse_trace}\nERROR: invalid value of the `unique` attribute. "
                f"Expected a boolean value. Get: {attr_conf['unique']}"
            )
        unique = attr_conf.get("unique", False)

        if "sorted" in attr_conf:
            Validator.must_in(
                attr_conf["sorted"],
                self.SORTED_VALUES,
                f"{parse_trace}\nParsing `sorted` of the attribute",
            )
        sorted = Sorted(attr_conf.get("sorted", Sorted.Null.value))

        if "value_type" in attr_conf:
            Validator.must_in(
                attr_conf["value_type"],
                self.VALUE_TYPE_VALUES,
                f"{parse_trace}\nParsing `value_type` of the attribute",
            )
        value_type = ValueType(
            attr_conf.get("value_type", ValueType.UnspecifiedSingle.value)
        )

        if "missing_values" in attr_conf:
            trace = f"{parse_trace}\nParsing missing_values of the attribute"
            Validator.must_be_list(attr_conf["missing_values"], trace)
            for val in attr_conf["missing_values"]:
                if val is not None and not isinstance(val, (str, int, bool, float)):
                    raise InputError(
                        f"{trace}\nERROR: invalid value. Expected either one of string, "
                        f"integer, or float. Get f{type(val)} instead"
                    )

        missing_values = attr_conf.get("missing_values", [])
        return Attr(
            attr_id, resource_id, path, missing_values, unique, sorted, value_type
        )



from __future__ import annotations

import os
from typing import Any, Optional

from drepr.models.sm import DREPR_URI
from drepr.utils.namespace_mixin import NamespaceManager
from drepr.writers.base import StreamClassWriter
from rdflib import RDF, BNode, Literal, URIRef

SubjVal = str | tuple | int | bool


class TurtleWriter(StreamClassWriter):
    def __init__(self, prefixes: dict[str, str], normalize_uri: Optional[bool] = None):
        if normalize_uri is None:
            normalize_uri = bool(
                int(os.environ.get("drepr.writer.turtle_writer.normalize_uri", "1"))
            )

        self.write_stream = []
        if normalize_uri:
            self.namespace_manager = NamespaceManager.from_prefix2ns(prefixes)
            for prefix in prefixes:
                self.write_stream.append(f"@prefix {prefix}: <{prefixes[prefix]}> .\n")
            self.write_stream.append("\n")
        else:
            self.namespace_manager = None
        self.written_records: dict[SubjVal, BNode | URIRef] = {}
        self.origin_subj: SubjVal = ""
        self.subj: Optional[URIRef | BNode] = None
        self.buffer: list[tuple[URIRef, URIRef | BNode | Literal]] = []
        self.is_buffered: bool = False
        # whether the current subject has any data or not
        # if the subject has URI, then it has data
        self.subj_has_data: bool = False

    def has_written_record(self, subj: SubjVal) -> bool:
        return subj in self.written_records

    def begin_record(
        self, class_uri: str, subj: SubjVal, is_blank: bool, is_buffered: bool
    ):
        self.origin_subj = subj
        if is_blank:
            if subj in self.written_records:
                self.subj = self.written_records[subj]
                self.subj_has_data = True
            else:
                self.subj = BNode()
                self.subj_has_data = False
        else:
            # subj will be a string for URIRef
            self.subj = URIRef(subj)  # type: ignore
            self.subj_has_data = True

        if is_buffered:
            self.buffer = [(RDF.type, URIRef(class_uri))]
        else:
            self.write_triple(self.subj, RDF.type, URIRef(class_uri))
        self.is_buffered = is_buffered
        self.subj_has_data = False

    def end_record(self):
        if self.subj is None:
            # has been aborted
            return

        if len(self.buffer) > 0:
            pred, obj = self.buffer[0]
            self.write_triple(self.subj, pred, obj)
            for i in range(1, len(self.buffer)):
                pred, obj = self.buffer[i]
                self.write_pred_obj(pred, obj)
            self.buffer = []
        self.write_stream[-1] = " .\n\n"
        self.written_records[self.origin_subj] = self.subj
        self.subj = None
        self.subj_has_data = False

    def abort_record(self):
        """Abort the record that is being written"""
        self.subj = None
        self.subj_has_data = False
        self.buffer = []

    def is_record_empty(self) -> bool:
        return not self.subj_has_data

    def write_data_property(self, predicate_id: str, value: Any, dtype: Optional[str]):
        if self.subj is None:
            return
        if dtype == DREPR_URI:
            value = URIRef(value)
        else:
            # to handle a bug in RDFlib that does not serialize integer properly.
            if (
                dtype == "http://www.w3.org/2001/XMLSchema#integer"
                or dtype == "http://www.w3.org/2001/XMLSchema#long"
                or dtype == "http://www.w3.org/2001/XMLSchema#int"
            ):
                value = int(float(value))
            value = Literal(value, datatype=dtype)

        self.subj_has_data = True
        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), value))
        else:
            assert self.subj is not None
            self.write_pred_obj(URIRef(predicate_id), value)

    def write_object_property(
        self,
        predicate_id: str,
        object: SubjVal,
        is_subject_blank: bool,
        is_object_blank: bool,
        is_new_subj: bool,
    ):
        if self.subj is None:
            return
        object = self.written_records[object]
        self.subj_has_data = True
        if self.is_buffered:
            self.buffer.append((URIRef(predicate_id), object))
        else:
            assert self.subj is not None
            self.write_pred_obj(URIRef(predicate_id), object)

    def write_to_string(self):
        return "".join(self.write_stream)

    def write_to_file(self, filepath):
        with open(filepath, "w") as f:
            for s in self.write_stream:
                f.write(s)

    def write_triple(self, subj: URIRef | BNode, pred, obj):
        self.write_stream.append(subj.n3(self.namespace_manager))  # type: ignore
        self.write_stream.append(" ")
        self.write_stream.append(pred.n3(self.namespace_manager))
        self.write_stream.append(" ")
        self.write_stream.append(obj.n3(self.namespace_manager))
        self.write_stream.append(" ;\n")

    def write_pred_obj(self, pred, obj):
        self.write_stream.append("\t")
        self.write_stream.append(pred.n3(self.namespace_manager))
        self.write_stream.append(" ")
        self.write_stream.append(obj.n3(self.namespace_manager))
        self.write_stream.append(" ;\n")



import re
from abc import ABC, abstractmethod
from copy import copy
from typing import Any, List, Optional, Union

from drepr.models.parsers.interface import PathParser
from drepr.models.path import Expr, IndexExpr, Path, RangeExpr, WildcardExpr
from drepr.models.resource import Resource, ResourceType
from drepr.utils.validator import InputError


class PathParserV2(PathParser):
    """
    Similar to path parser v1. However, we now allow special using indexing name per resource type

    For spreadsheet, we allow column to be letter
    """

    REG_SRANGE = re.compile(r"^(\d+)?\.\.(-?\d+)?(?::(\d+))?$")
    REG_SINDEX = re.compile(r"^(?:\$\{([^}]+)})|(\d+)|(.*)$")
    REG_SRANGE_EXPR = re.compile(
        r"^(?:(\d+)|(?:\$\{([^}]+)}))?\.\.(?:(-\d+)|(?:\$\{([^}]+)}))?(?::(\d+)|(?:\$\{([^}]+)}))?$"
    )

    REG_JPATH_BRACKET = re.compile(
        r"(?:\[(-?\d+)?\:(-?\d+)?(?:\:(-?\d+))?\])|(?:\[(-?\d+)\])|(?:\['([^']+)'\])"
    )
    REG_JPATH_DOT = re.compile(r"\.((?:(?!\.|\[).)+)")

    def parse(self, resource: Optional[Resource], path: Any, parse_trace: str) -> Path:
        if isinstance(path, str):
            return self.parse_jsonpath(resource, path, parse_trace)

        if isinstance(path, list):
            return self.parse_custom_path(resource, path, parse_trace)

        raise InputError(
            f"{parse_trace}\nERROR: the path must either be a "
            f"string (JSONPath) or a list of steps. Get {type(path)} instead"
        )

    # noinspection PyMethodMayBeStatic
    def letter2index(self, letter: str) -> int:
        letter = letter.lower()
        n_chars = ord("z") - ord("a") + 1
        index = 0
        for i, c in enumerate(reversed(letter)):
            assert (
                ord("a") <= ord(c) <= ord("z")
            ), f"{c} is not a valid column in spreadsheet"
            index += (ord(c) - ord("a") + 1) * (n_chars**i)
        return index - 1

    def isdigit(self, s: str) -> bool:
        if s.startswith("-"):
            return s[1:].isdigit()
        return s.isdigit()

    def parse_jsonpath(
        self, resource: Optional[Resource], jpath: str, parse_trace: str
    ) -> Path:
        if not jpath.startswith("$"):
            raise InputError(
                f"{parse_trace}\nERROR: invalid json path. The path must start with `$`. "
                f"Get: {jpath}"
            )

        jpath = jpath[1:]
        steps = []
        parsing_pos = 1

        # pre-processing the spreadsheet resource to allow letter column
        if resource is not None:
            if resource.type == ResourceType.Spreadsheet:
                last_step_index = max(jpath.rfind("["), jpath.rfind("."))
                if jpath[last_step_index] == "[":
                    last_step = jpath[last_step_index + 1 : -1]
                    result = last_step.split(":")
                    if len(result) == 1:
                        index = result[0]
                        if not self.isdigit(index):
                            new_last_step = self.letter2index(index)
                        else:
                            new_last_step = index
                    else:
                        if len(result) == 3:
                            start, end, step = result
                        elif len(result) == 2:
                            start, end = result
                            step = 1
                        else:
                            raise InputError(f"{parse_trace}\nERROR: invalid path")

                        if (len(start) > 0 and not self.isdigit(start)) or (
                            len(end) > 0 and not self.isdigit(end)
                        ):
                            # they use letter system, otherwise, do nothing
                            if (len(start) > 0 and self.isdigit(start)) or (
                                len(end) > 0 and self.isdigit(end)
                            ):
                                raise InputError(
                                    f"{parse_trace}\nERROR: Cannot mixed between number and letter index"
                                )
                            start = self.letter2index(start)
                            if len(end) > 0:
                                end = self.letter2index(end)
                            new_last_step = f"{start}:{end}:{step}"
                        else:
                            new_last_step = f"{start}:{end}:{step}"

                    jpath = jpath[:last_step_index] + f"[{new_last_step}]"
                elif jpath[last_step_index] == ".":
                    last_step = jpath[last_step_index + 1 :]
                    if not self.isdigit(last_step):
                        new_last_step = self.letter2index(last_step)
                    else:
                        new_last_step = last_step
                    jpath = jpath[:last_step_index] + f".{new_last_step}"

        while len(jpath) > 0:
            if jpath.startswith("["):
                m = self.REG_JPATH_BRACKET.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing bracket at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                if m.group(5) is not None:
                    # match with string
                    steps.append(IndexExpr(m.group(5)))
                elif m.group(4) is not None:
                    # match with a single number
                    steps.append(IndexExpr(int(m.group(4))))
                else:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
            elif jpath.startswith(".*~"):
                # *~ select property names
                steps.append(WildcardExpr.Names)
                jpath = jpath[3:]
                parsing_pos += 3
            elif jpath.startswith(".*"):
                steps.append(WildcardExpr.Values)
                jpath = jpath[2:]
                parsing_pos += 2
            else:
                m = self.REG_JPATH_DOT.match(jpath)
                if m is None:
                    raise InputError(
                        f"{parse_trace}\nERROR: invalid json path, error while parsing step at position {parsing_pos}"
                    )

                jpath = jpath[m.span()[-1] :]
                parsing_pos += m.span()[-1]  # m.span()[0] is always 0

                # after a dot, it can either be a number or a string
                if m.group(1).isdigit():
                    steps.append(IndexExpr(int(m.group(1))))
                else:
                    steps.append(IndexExpr(m.group(1)))

        return Path(steps)

    def parse_custom_path(
        self, resource: Optional[Resource], path: List[str], parse_trace: str
    ) -> Path:
        if resource is not None:
            if resource.type == ResourceType.Spreadsheet:
                path = copy(path)
                last_step = path[-1]
                if isinstance(last_step, str):
                    if last_step.find("..") != -1:
                        tmp = last_step.split(":")
                        start, end = tmp[0].split("..")
                        if len(tmp) == 2:
                            step = f":{tmp[1]}"
                        else:
                            step = ""

                        if (len(start) > 0 and not self.isdigit(start)) or (
                            len(end) > 0 and not self.isdigit(end)
                        ):
                            # they use letter system, otherwise, do nothing
                            if (len(start) > 0 and self.isdigit(start)) or (
                                len(end) > 0 and self.isdigit(end)
                            ):
                                raise InputError(
                                    f"{parse_trace}\nERROR: Cannot mixed between number and letter index"
                                )
                            start = self.letter2index(start)
                            if len(end) > 0:
                                end = self.letter2index(end)
                            new_last_step = f"{start}..{end}{step}"
                        else:
                            new_last_step = f"{start}..{end}{step}"
                        path[-1] = new_last_step
                    elif not self.isdigit(last_step):
                        path[-1] = self.letter2index(last_step)

        steps = []
        for i, step in enumerate(path):
            trace = f"Parsing step {i} ({step})"
            if isinstance(step, str):
                m = self.REG_SRANGE.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            int(m.group(1) or "0"),
                            int(m.group(2)) if m.group(2) is not None else None,
                            int(m.group(3) or "1"),
                        )
                    )
                    continue

                m = self.REG_SRANGE_EXPR.match(step)
                if m is not None:
                    steps.append(
                        RangeExpr(
                            (
                                (
                                    Expr(m.group(1)[2:-1])
                                    if m.group(1).startswith("${")
                                    else int(m.group(1))
                                )
                                if m.group(1) is not None
                                else 0
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else None
                            ),
                            (
                                (
                                    Expr(m.group(2)[2:-1])
                                    if m.group(2).startswith("${")
                                    else int(m.group(2))
                                )
                                if m.group(2) is not None
                                else 1
                            ),
                        )
                    )
                    continue

                if step.startswith("${"):
                    steps.append(IndexExpr(Expr(step[2:-1])))
                else:
                    steps.append(IndexExpr(step))
            elif isinstance(step, int):
                steps.append(IndexExpr(step))
            elif isinstance(step, list):
                assert (
                    len(step) == 1
                ), "The list notation [...] is used to annotate optional index, so the list must have only one element"
                steps.append(IndexExpr(step[0], is_optional=True))
            else:
                raise InputError(
                    f"{parse_trace}\n{trace}\nERROR: step must either be string or number. Get {type(step)} instead"
                )

        return Path(steps)
